# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from tests import HTTPResponse


class TestDeleteRoleAssociationsCommand:
    def assert_list_call_matches(
        self, aws_request, service_account, cluster_name, namespace
    ):
        assert aws_request.service_name == "eks"
        assert aws_request.operation_name == "ListPodIdentityAssociations"
        assert aws_request.params["clusterName"] == cluster_name
        assert aws_request.params["namespace"] == namespace
        assert aws_request.params["serviceAccount"] == service_account

    def assert_delete_call_matches(
        self, aws_request, cluster_name, association_id
    ):
        assert aws_request.service_name == "eks"
        assert aws_request.operation_name == "DeletePodIdentityAssociation"
        assert aws_request.params["clusterName"] == cluster_name
        assert aws_request.params["associationId"] == association_id

    # Use case: Expect to return delete pod identity association results for start job run
    # Expected results: Operation is performed by client
    # to delete pod identity associations with start job run service accounts
    def test_delete_role_associations_for_start_job_run(
        self,
        cli_runner,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        list_pod_identity_association_response,
        delete_pod_identity_association_response,
        start_job_run_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        for i in range(3):
            cli_runner.add_response(
                HTTPResponse(body=list_pod_identity_association_response[i])
            )
            cli_runner.add_response(
                HTTPResponse(body=delete_pod_identity_association_response[i])
            )

        result = cli_runner.run(
            [
                "emr-containers",
                "delete-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
            ]
        )

        request_idx = 0
        assert result.aws_requests[request_idx].service_name == "eks"
        assert (
            result.aws_requests[request_idx].operation_name
            == "DescribeCluster"
        )
        assert (
            len(result.aws_requests)
            == (len(start_job_run_service_accounts) * 2) + 1
        )
        for i in range(len(start_job_run_service_accounts)):
            request_idx += 1
            self.assert_list_call_matches(
                result.aws_requests[request_idx],
                start_job_run_service_accounts[i],
                cluster_name,
                namespace,
            )
            request_idx += 1
            self.assert_delete_call_matches(
                result.aws_requests[request_idx], cluster_name, str(i + 1)
            )

    # Use case: Expect to return delete pod identity association results for interactive endpoint
    # Expected results: Operation is performed by client
    # to delete pod identity associations with interactive endpoint service accounts
    def test_delete_role_associations_for_interactive_endpoint(
        self,
        cli_runner,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        list_pod_identity_association_response,
        delete_pod_identity_association_response,
        interactive_endpoint_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        for i in range(3):
            cli_runner.add_response(
                HTTPResponse(body=list_pod_identity_association_response[i])
            )
            cli_runner.add_response(
                HTTPResponse(body=delete_pod_identity_association_response[i])
            )

        result = cli_runner.run(
            [
                "emr-containers",
                "delete-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "interactive_endpoint",
            ]
        )

        request_idx = 0
        assert result.aws_requests[request_idx].service_name == "eks"
        assert (
            result.aws_requests[request_idx].operation_name
            == "DescribeCluster"
        )
        assert (
            len(result.aws_requests)
            == (len(interactive_endpoint_service_accounts) * 2) + 1
        )
        for i in range(len(interactive_endpoint_service_accounts)):
            request_idx += 1
            self.assert_list_call_matches(
                result.aws_requests[request_idx],
                interactive_endpoint_service_accounts[i],
                cluster_name,
                namespace,
            )
            request_idx += 1
            self.assert_delete_call_matches(
                result.aws_requests[request_idx], cluster_name, str(i + 1)
            )

    # Use case: Expect to return delete pod identity association results for spark operator
    # Expected results: Operation is performed by client
    # to delete pod identity associations with spark operator service accounts
    def test_delete_role_associations_for_spark_operator(
        self,
        cli_runner,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        list_pod_identity_association_response,
        delete_pod_identity_association_response,
        spark_operator_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        for i in range(3):
            cli_runner.add_response(
                HTTPResponse(body=list_pod_identity_association_response[i])
            )
            cli_runner.add_response(
                HTTPResponse(body=delete_pod_identity_association_response[i])
            )

        result = cli_runner.run(
            [
                "emr-containers",
                "delete-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "spark_operator",
            ]
        )

        request_idx = 0
        assert result.aws_requests[request_idx].service_name == "eks"
        assert (
            result.aws_requests[request_idx].operation_name
            == "DescribeCluster"
        )
        assert (
            len(result.aws_requests)
            == (len(spark_operator_service_accounts) * 2) + 1
        )
        for i in range(len(spark_operator_service_accounts)):
            request_idx += 1
            self.assert_list_call_matches(
                result.aws_requests[request_idx],
                spark_operator_service_accounts[i],
                cluster_name,
                namespace,
            )
            request_idx += 1
            self.assert_delete_call_matches(
                result.aws_requests[request_idx], cluster_name, str(i + 1)
            )

    # Use case: Expect to return delete pod identity association results for spark operator with operator namespace
    # Expected results: Operation is performed by client
    # to delete pod identity associations with spark operator service accounts in correct namespaces
    def test_delete_role_associations_for_spark_operator_namespace(
        self,
        cli_runner,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        list_pod_identity_association_response,
        delete_pod_identity_association_response,
        spark_operator_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        for i in range(3):
            cli_runner.add_response(
                HTTPResponse(body=list_pod_identity_association_response[i])
            )
            cli_runner.add_response(
                HTTPResponse(body=delete_pod_identity_association_response[i])
            )

        result = cli_runner.run(
            [
                "emr-containers",
                "delete-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "spark_operator",
                "--operator-namespace",
                "spark-operator",
            ]
        )

        request_idx = 0
        assert result.aws_requests[request_idx].service_name == "eks"
        assert (
            result.aws_requests[request_idx].operation_name
            == "DescribeCluster"
        )
        assert (
            len(result.aws_requests)
            == (len(spark_operator_service_accounts) * 2) + 1
        )
        for i in range(len(spark_operator_service_accounts)):
            request_idx += 1
            ns = "spark-operator" if i == 0 else namespace
            self.assert_list_call_matches(
                result.aws_requests[request_idx],
                spark_operator_service_accounts[i],
                cluster_name,
                ns,
            )
            request_idx += 1
            self.assert_delete_call_matches(
                result.aws_requests[request_idx], cluster_name, str(i + 1)
            )

    # Use case: Expect to return delete pod identity association results for flink operator
    # Expected results: Operation is performed by client
    # to delete pod identity associations with flink operator service accounts
    def test_delete_role_associations_for_flink_operator(
        self,
        cli_runner,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        list_pod_identity_association_response,
        delete_pod_identity_association_response,
        flink_operator_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        for i in range(3):
            cli_runner.add_response(
                HTTPResponse(body=list_pod_identity_association_response[i])
            )
            cli_runner.add_response(
                HTTPResponse(body=delete_pod_identity_association_response[i])
            )

        result = cli_runner.run(
            [
                "emr-containers",
                "delete-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "flink_operator",
            ]
        )

        request_idx = 0
        assert result.aws_requests[request_idx].service_name == "eks"
        assert (
            result.aws_requests[request_idx].operation_name
            == "DescribeCluster"
        )
        assert (
            len(result.aws_requests)
            == (len(flink_operator_service_accounts) * 2) + 1
        )
        for i in range(len(flink_operator_service_accounts)):
            request_idx += 1
            self.assert_list_call_matches(
                result.aws_requests[request_idx],
                flink_operator_service_accounts[i],
                cluster_name,
                namespace,
            )
            request_idx += 1
            self.assert_delete_call_matches(
                result.aws_requests[request_idx], cluster_name, str(i + 1)
            )

    # Use case: Expect to return delete pod identity association results for flink operator with operator namespaces
    # Expected results: Operation is performed by client
    # to delete pod identity associations with flink operator service accounts in correct namespace
    def test_delete_role_associations_for_flink_operator_namespace(
        self,
        cli_runner,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        list_pod_identity_association_response,
        delete_pod_identity_association_response,
        flink_operator_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        for i in range(3):
            cli_runner.add_response(
                HTTPResponse(body=list_pod_identity_association_response[i])
            )
            cli_runner.add_response(
                HTTPResponse(body=delete_pod_identity_association_response[i])
            )

        result = cli_runner.run(
            [
                "emr-containers",
                "delete-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "flink_operator",
                "--operator-namespace",
                "flink-operator",
            ]
        )

        request_idx = 0
        assert result.aws_requests[request_idx].service_name == "eks"
        assert (
            result.aws_requests[request_idx].operation_name
            == "DescribeCluster"
        )
        assert (
            len(result.aws_requests)
            == (len(flink_operator_service_accounts) * 2) + 1
        )
        for i in range(len(flink_operator_service_accounts)):
            request_idx += 1
            ns = "flink-operator" if i == 0 else namespace
            self.assert_list_call_matches(
                result.aws_requests[request_idx],
                flink_operator_service_accounts[i],
                cluster_name,
                ns,
            )
            request_idx += 1
            self.assert_delete_call_matches(
                result.aws_requests[request_idx], cluster_name, str(i + 1)
            )

    # Use case: Expect to return delete pod identity association results for livy
    # Expected results: Operation is performed by client
    # to delete pod identity associations with livy service accounts
    def test_delete_role_associations_for_livy(
        self,
        cli_runner,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        list_pod_identity_association_response,
        delete_pod_identity_association_response,
        livy_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        for i in range(2):
            cli_runner.add_response(
                HTTPResponse(body=list_pod_identity_association_response[i])
            )
            cli_runner.add_response(
                HTTPResponse(body=delete_pod_identity_association_response[i])
            )

        result = cli_runner.run(
            [
                "emr-containers",
                "delete-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "livy",
            ]
        )

        request_idx = 0
        assert result.aws_requests[request_idx].service_name == "eks"
        assert (
            result.aws_requests[request_idx].operation_name
            == "DescribeCluster"
        )
        assert len(result.aws_requests) == (len(livy_service_accounts) * 2) + 1
        for i in range(len(livy_service_accounts)):
            request_idx += 1
            self.assert_list_call_matches(
                result.aws_requests[request_idx],
                livy_service_accounts[i],
                cluster_name,
                namespace,
            )
            request_idx += 1
            self.assert_delete_call_matches(
                result.aws_requests[request_idx], cluster_name, str(i + 1)
            )

    # Use case: Expect to return delete pod identity association results for livy with controller namespace
    # Expected results: Operation is performed by client
    # to delete pod identity associations with livy service accounts in correct namespace
    def test_delete_role_associations_for_livy_namespace(
        self,
        cli_runner,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        list_pod_identity_association_response,
        delete_pod_identity_association_response,
        livy_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        for i in range(2):
            cli_runner.add_response(
                HTTPResponse(body=list_pod_identity_association_response[i])
            )
            cli_runner.add_response(
                HTTPResponse(body=delete_pod_identity_association_response[i])
            )

        result = cli_runner.run(
            [
                "emr-containers",
                "delete-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "livy",
                "--operator-namespace",
                "livy",
            ]
        )

        request_idx = 0
        assert result.aws_requests[request_idx].service_name == "eks"
        assert (
            result.aws_requests[request_idx].operation_name
            == "DescribeCluster"
        )
        assert len(result.aws_requests) == (len(livy_service_accounts) * 2) + 1
        for i in range(len(livy_service_accounts)):
            request_idx += 1
            ns = "livy" if i == 0 else namespace
            self.assert_list_call_matches(
                result.aws_requests[request_idx],
                livy_service_accounts[i],
                cluster_name,
                ns,
            )
            request_idx += 1
            self.assert_delete_call_matches(
                result.aws_requests[request_idx], cluster_name, str(i + 1)
            )

    # Use case: Expect to return delete pod identity association results for customer input service accounts
    # Expected results: Operation is performed by client
    # to delete pod identity associations with customer input service accounts
    def test_delete_role_associations_for_customer_service_account(
        self,
        cli_runner,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        list_pod_identity_association_response,
        delete_pod_identity_association_response,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(
            HTTPResponse(body=list_pod_identity_association_response[0])
        )
        cli_runner.add_response(
            HTTPResponse(body=delete_pod_identity_association_response[0])
        )

        result = cli_runner.run(
            [
                "emr-containers",
                "delete-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--service-account-name",
                "test_sa",
            ]
        )

        assert result.aws_requests[0].service_name == "eks"
        assert result.aws_requests[0].operation_name == "DescribeCluster"
        assert len(result.aws_requests) == 3
        self.assert_list_call_matches(
            result.aws_requests[1], "test_sa", cluster_name, namespace
        )
        self.assert_delete_call_matches(
            result.aws_requests[2], cluster_name, "1"
        )

    # Use case: Expect to do nothing on deletion when resource not found but print warning message
    # Expected results: Association deletions are skipped
    def test_delete_role_associations_already_exists_for_start_job_run(
        self,
        cli_runner,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        start_job_run_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(HTTPResponse(body=[]))
        cli_runner.add_response(HTTPResponse(body=[]))
        cli_runner.add_response(HTTPResponse(body=[]))
        result = cli_runner.run(
            [
                "emr-containers",
                "delete-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
            ]
        )

        expected_out = ""
        for service_account in start_job_run_service_accounts:
            expected_out += (
                f"Skipping deletion as no pod identity association found for service account {service_account} "
                f"and role {role_name} in namespace {namespace}\n"
            )
        assert result.stderr == expected_out



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.s3.filegenerator import FileStat
from awscli.customizations.s3.fileinfo import FileInfo
from awscli.customizations.s3.fileinfobuilder import FileInfoBuilder
from awscli.testutils import unittest


class TestFileInfoBuilder(unittest.TestCase):
    def test_info_setter(self):
        info_setter = FileInfoBuilder(
            client='client',
            source_client='source_client',
            parameters='parameters',
            is_stream='is_stream',
        )
        files = [
            FileStat(
                src='src',
                dest='dest',
                compare_key='compare_key',
                size='size',
                last_update='last_update',
                src_type='src_type',
                dest_type='dest_type',
                operation_name='operation_name',
                response_data='associated_response_data',
                etag='etag',
                case_conflict_submitted='case_conflict_submitted',
                case_conflict_key='case_conflict_key',
            )
        ]
        file_infos = info_setter.call(files)
        for file_info in file_infos:
            attributes = file_info.__dict__.keys()
            for key in attributes:
                self.assertEqual(getattr(file_info, key), str(key))

    def test_swaps_clients_for_sync_delete(self):
        client_name = 'client'
        source_client_name = 'source_client'
        info_setter = FileInfoBuilder(
            client=client_name,
            source_client=source_client_name,
            parameters={'delete': True},
            is_stream='is_stream',
        )
        files = [
            FileStat(
                src='src',
                dest='dest',
                compare_key='compare_key',
                size='size',
                last_update='last_update',
                src_type='src_type',
                dest_type='dest_type',
                operation_name='delete',
            )
        ]
        file_infos = info_setter.call(files)
        for file_info in file_infos:
            self.assertEqual(file_info.client, source_client_name)
            self.assertEqual(file_info.source_client, client_name)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import platform

import pytest
from prompt_toolkit.application import create_app_session
from prompt_toolkit.input import create_pipe_input
from prompt_toolkit.output import DummyOutput

import awscli.logger


@pytest.fixture(autouse=True)
def clear_loggers():
    """Ensure all loggers have no residual state before test runs

    Some tests rely on updating the built-in logger and so we want to make
    sure that any residual state is cleared between test cases such as making
    sure there are no handlers and the logger level is reset to not set.
    """
    loggers = [name for name in logging.root.manager.loggerDict]
    for logger_name in loggers:
        logger = logging.getLogger(logger_name)
        logger.handlers = []
        logger.setLevel(logging.NOTSET)
    awscli.logger.disable_crt_logging()


@pytest.fixture
def ptk_app_session():
    with create_pipe_input() as pipe_input:
        output = DummyOutput()
        try:
            with create_app_session(
                input=pipe_input, output=output
            ) as session:
                yield session
        finally:
            pipe_input.close()



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestPutAccountDetails(BaseAWSCommandParamsTest):
    prefix = 'sesv2 put-account-details'

    def test_put_account_details(self):
        cmdline = self.prefix
        cmdline += '  --website-url https://example.com '
        cmdline += '--mail-type MARKETING '
        cmdline += '--use-case-description Description'

        result = {
            'WebsiteURL': 'https://example.com',
            'MailType': 'MARKETING',
            'UseCaseDescription': 'Description',
        }
        self.assert_params_for_cmd(cmdline, result)



import json
import os
import subprocess
import sys

RST_GENERATION_SCRIPT = 'htmlgen'
script_path = os.path.join(os.path.dirname(__file__), RST_GENERATION_SCRIPT)
os.environ['PATH'] += ':.'
rc = subprocess.call("python " + script_path, shell=True, env=os.environ)
if rc != 0:
    sys.stderr.write("Failed to generate documentation!\n")
    sys.exit(2)



# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import importlib.metadata
import re
import sys
from pathlib import Path

from constants import (
    BOOTSTRAP_REQUIREMENTS,
    PORTABLE_EXE_REQUIREMENTS,
)
from utils import (
    UnmetDependenciesException,
    get_install_requires,
    parse_requirements,
)

ROOT = Path(__file__).parents[2]
PYPROJECT = ROOT / "pyproject.toml"
BUILD_REQS_RE = re.compile(r"requires = \[([\s\S]+?)\]\s", re.MULTILINE)
EXTRACT_DEPENDENCIES_RE = re.compile(r'"(.+)"')


def validate_env(target_artifact):
    requirements = _get_requires_list(target_artifact)
    unmet_deps = _get_unmet_dependencies(requirements)
    if unmet_deps:
        in_venv = sys.prefix != sys.base_prefix
        raise UnmetDependenciesException(unmet_deps, in_venv)


def _get_requires_list(target_artifact):
    requires_list = _parse_pyproject_requirements()
    requires_list += _parse_requirements(BOOTSTRAP_REQUIREMENTS)
    requires_list += get_install_requires()
    if target_artifact == "portable-exe":
        requires_list += _parse_requirements(PORTABLE_EXE_REQUIREMENTS)
    return parse_requirements(requires_list)


def _parse_pyproject_requirements():
    with open(PYPROJECT) as f:
        data = f.read()
    raw_dependencies = BUILD_REQS_RE.findall(data)[0]
    dependencies = EXTRACT_DEPENDENCIES_RE.findall(raw_dependencies)
    return list(dependencies)


def _parse_requirements(requirements_file):
    requirements = []
    with open(requirements_file) as f:
        for line in f.readlines():
            if not line.startswith(("-r", "#")):
                requirements.append(line.strip())
    return requirements


def _get_unmet_dependencies(requirements):
    unmet = []
    checked = set()
    for requirement in requirements:
        project_name = requirement.name
        if project_name in checked:
            continue
        checked.add(project_name)
        try:
            actual_version = importlib.metadata.version(project_name)
        except importlib.metadata.PackageNotFoundError:
            try:
                # Packages built from source may have directory names
                # that replace "." with "_".
                actual_version = importlib.metadata.version(
                    project_name.replace(".", "_")
                )
            except importlib.metadata.PackageNotFoundError:
                unmet.append((project_name, None, requirement))
                continue
        if not requirement.is_in_range(actual_version):
            unmet.append((project_name, actual_version, requirement))
    return unmet



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import contextlib
import io
import logging

from prompt_toolkit.buffer import Buffer

from awscli.autoprompt.logger import PromptToolkitHandler
from awscli.testutils import mock, unittest


class TestPromptToolkitHandler(unittest.TestCase):
    @mock.patch('awscli.autoprompt.logger.get_app')
    def test_can_log_to_prompter(self, get_app):
        handler = PromptToolkitHandler()
        handler.setLevel(logging.DEBUG)
        log = logging.getLogger('test_prompt_logger')
        log.setLevel(logging.DEBUG)
        log.handlers = []
        log.addHandler(handler)
        buffer = Buffer()
        app = mock.Mock()
        app.debug = True
        app.is_running = True
        app.layout.get_buffer_by_name.return_value = buffer
        get_app.return_value = app
        log.debug('error message')
        self.assertIn('error message', buffer.document.text)

    def test_log_to_stderr_if_no_app(self):
        fake_stderr = io.StringIO()
        with contextlib.redirect_stderr(fake_stderr):
            handler = PromptToolkitHandler()
            handler.setLevel(logging.DEBUG)
            log = logging.getLogger('test_prompt_logger')
            log.handlers = []
            log.addHandler(handler)
            log.setLevel(logging.DEBUG)
            log.debug('error message')
        self.assertIn('error message', fake_stderr.getvalue())



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from decimal import Decimal

from awscli.customizations.dynamodb.exceptions import (
    EmptyExpressionError,
    InvalidLiteralValueError,
    UnexpectedTokenError,
    UnknownExpressionError,
)
from awscli.customizations.dynamodb.parser import Parser
from awscli.testutils import unittest


class FakeLexer:
    def __init__(self, tokens):
        self._tokens = tokens

    def tokenize(self, expression):
        return self._tokens


class TestParser(unittest.TestCase):
    maxDiff = None

    def assert_parse(self, tokens, expected):
        self.assertEqual(expected, self._parse(tokens))

    def _parse(self, tokens, expression=None):
        if expression is None:
            self._insert_token_positions(tokens)
            expression = ''
            for token in tokens:
                expression += str(token['value']) + ' '
            expression = expression.strip()
        lexer = FakeLexer(tokens)
        parser = Parser(lexer)
        return parser.parse(expression)

    def _insert_token_positions(self, tokens):
        position = 0
        for token in tokens:
            if 'start' not in token:
                token['start'] = position
            if 'end' not in token:
                token['end'] = token['start'] + len(str(token['value']))
            position = token['end'] + 1

    def test_parse_sequence(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'comma', 'value': ','},
            {'type': 'identifier', 'value': 'bar'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'sequence',
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
                {'type': 'identifier', 'value': 'bar', 'children': []},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_single_identifier(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {'type': 'identifier', 'value': 'foo', 'children': []}
        self.assert_parse(tokens, expected)

    def test_parse_and_expression(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'and', 'value': 'and'},
            {'type': 'identifier', 'value': 'baz'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bam'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'and_expression',
            'children': [
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                        {'type': 'literal', 'value': 'bar', 'children': []},
                    ],
                },
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'baz', 'children': []},
                        {'type': 'literal', 'value': 'bam', 'children': []},
                    ],
                },
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_multiple_and_expressions(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'and', 'value': 'and'},
            {'type': 'identifier', 'value': 'baz'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bam'},
            {'type': 'and', 'value': 'and'},
            {'type': 'identifier', 'value': 'spam'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'eggs'},
            {'type': 'eof', 'value': ''},
        ]
        left_and_expression = {
            'type': 'and_expression',
            'children': [
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                        {'type': 'literal', 'value': 'bar', 'children': []},
                    ],
                },
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'baz', 'children': []},
                        {'type': 'literal', 'value': 'bam', 'children': []},
                    ],
                },
            ],
        }
        expected = {
            'type': 'and_expression',
            'children': [
                left_and_expression,
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {
                            'type': 'identifier',
                            'value': 'spam',
                            'children': [],
                        },
                        {'type': 'literal', 'value': 'eggs', 'children': []},
                    ],
                },
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_and_missing_second_expression(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'and', 'value': 'and'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnknownExpressionError):
            self._parse(tokens)

    def test_parse_or_expression(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'or', 'value': 'or'},
            {'type': 'identifier', 'value': 'baz'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bam'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'or_expression',
            'children': [
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                        {'type': 'literal', 'value': 'bar', 'children': []},
                    ],
                },
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'baz', 'children': []},
                        {'type': 'literal', 'value': 'bam', 'children': []},
                    ],
                },
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_multiple_or_expressions(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'or', 'value': 'or'},
            {'type': 'identifier', 'value': 'baz'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bam'},
            {'type': 'or', 'value': 'or'},
            {'type': 'identifier', 'value': 'spam'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'eggs'},
            {'type': 'eof', 'value': ''},
        ]
        left_or_expression = {
            'type': 'or_expression',
            'children': [
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                        {'type': 'literal', 'value': 'bar', 'children': []},
                    ],
                },
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'baz', 'children': []},
                        {'type': 'literal', 'value': 'bam', 'children': []},
                    ],
                },
            ],
        }
        expected = {
            'type': 'or_expression',
            'children': [
                left_or_expression,
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {
                            'type': 'identifier',
                            'value': 'spam',
                            'children': [],
                        },
                        {'type': 'literal', 'value': 'eggs', 'children': []},
                    ],
                },
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_or_missing_second_expression(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'or', 'value': 'or'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnknownExpressionError):
            self._parse(tokens)

    def test_parse_mixed_and_or_expressions(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'and', 'value': 'and'},
            {'type': 'identifier', 'value': 'baz'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bam'},
            {'type': 'or', 'value': 'or'},
            {'type': 'identifier', 'value': 'spam'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'eggs'},
            {'type': 'eof', 'value': ''},
        ]
        left_and_expression = {
            'type': 'and_expression',
            'children': [
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                        {'type': 'literal', 'value': 'bar', 'children': []},
                    ],
                },
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'baz', 'children': []},
                        {'type': 'literal', 'value': 'bam', 'children': []},
                    ],
                },
            ],
        }
        expected = {
            'type': 'or_expression',
            'children': [
                left_and_expression,
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {
                            'type': 'identifier',
                            'value': 'spam',
                            'children': [],
                        },
                        {'type': 'literal', 'value': 'eggs', 'children': []},
                    ],
                },
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_subexpression(self):
        tokens = [
            {'type': 'lparen', 'value': '('},
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'rparen', 'value': ')'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'subexpression',
            'children': [
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                        {'type': 'literal', 'value': 'bar', 'children': []},
                    ],
                },
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_subexpression_with_and(self):
        tokens = [
            {'type': 'lparen', 'value': '('},
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'and', 'value': 'and'},
            {'type': 'identifier', 'value': 'baz'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bam'},
            {'type': 'rparen', 'value': ')'},
            {'type': 'eof', 'value': ''},
        ]
        and_expression = {
            'type': 'and_expression',
            'children': [
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                        {'type': 'literal', 'value': 'bar', 'children': []},
                    ],
                },
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'baz', 'children': []},
                        {'type': 'literal', 'value': 'bam', 'children': []},
                    ],
                },
            ],
        }
        expected = {'type': 'subexpression', 'children': [and_expression]}
        self.assert_parse(tokens, expected)

    def test_parse_subexpression_unmatched_paren(self):
        tokens = [
            {'type': 'lparen', 'value': '('},
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_parse_not_expression(self):
        tokens = [
            {'type': 'not', 'value': 'not'},
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'not_expression',
            'children': [
                {
                    'type': 'comparator',
                    'value': 'eq',
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                        {'type': 'literal', 'value': 'bar', 'children': []},
                    ],
                },
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_not_missing_expression(self):
        tokens = [
            {'type': 'not', 'value': 'not'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnknownExpressionError):
            self._parse(tokens)

    def test_parse_function(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'lparen', 'value': '('},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'comma', 'value': ','},
            {'type': 'literal', 'value': 'baz'},
            {'type': 'rparen', 'value': ')'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'function',
            'value': 'foo',
            'children': [
                {'type': 'literal', 'value': 'bar', 'children': []},
                {'type': 'literal', 'value': 'baz', 'children': []},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_function_missing_closing_paren(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'lparen', 'value': '('},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_function_args_missing_comma(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'lparen', 'value': '('},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'literal', 'value': 'baz'},
            {'type': 'rparen', 'value': ')'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_parse_in_expression(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'in', 'value': 'in'},
            {'type': 'lparen', 'value': '('},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'comma', 'value': ','},
            {'type': 'literal', 'value': 'baz'},
            {'type': 'rparen', 'value': ')'},
            {'type': 'eof', 'value': ''},
        ]
        sequence = {
            'type': 'sequence',
            'children': [
                {'type': 'literal', 'value': 'bar', 'children': []},
                {'type': 'literal', 'value': 'baz', 'children': []},
            ],
        }
        expected = {
            'type': 'in_expression',
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
                sequence,
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_in_missing_operand(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'in', 'value': 'in'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_in_missing_parens(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'in', 'value': 'in'},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'comma', 'value': ','},
            {'type': 'literal', 'value': 'baz'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_parse_between_expression(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'between', 'value': 'between'},
            {'type': 'literal', 'value': 1},
            {'type': 'and', 'value': 'and'},
            {'type': 'literal', 'value': 3},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'between_expression',
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
                {'type': 'literal', 'value': 1, 'children': []},
                {'type': 'literal', 'value': 3, 'children': []},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_between_missing_and(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'between', 'value': 'between'},
            {'type': 'literal', 'value': 1},
            {'type': 'literal', 'value': 3},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_parse_between_missing_right_operand(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'between', 'value': 'between'},
            {'type': 'literal', 'value': 1},
            {'type': 'and', 'value': 'and'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_parse_between_missing_center_operand(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'between', 'value': 'between'},
            {'type': 'and', 'value': 'and'},
            {'type': 'literal', 'value': 3},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_parse_comparison_expression(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'lte', 'value': '<='},
            {'type': 'literal', 'value': 8},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'comparator',
            'value': 'lte',
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
                {'type': 'literal', 'value': 8, 'children': []},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_unmatched_comparator(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'gte', 'value': '>='},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_parse_list_literal(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbracket', 'value': '['},
            {'type': 'literal', 'value': 8},
            {'type': 'comma', 'value': ','},
            {'type': 'literal', 'value': 9},
            {'type': 'rbracket', 'value': ']'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'comparator',
            'value': 'eq',
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
                {'type': 'literal', 'children': [], 'value': [8, 9]},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_empty_list(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbracket', 'value': '['},
            {'type': 'rbracket', 'value': ']'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'comparator',
            'value': 'eq',
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
                {'type': 'literal', 'children': [], 'value': []},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_list_literal_can_only_contain_literals(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbracket', 'value': '['},
            {'type': 'identifier', 'value': 'bar'},
            {'type': 'comma', 'value': ','},
            {'type': 'literal', 'value': 9},
            {'type': 'rbracket', 'value': ']'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_list_with_unmatched_bracket(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbracket', 'value': '['},
            {'type': 'literal', 'value': 9},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_parse_set_literal(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbrace', 'value': '{'},
            {'type': 'literal', 'value': Decimal('8')},
            {'type': 'comma', 'value': ','},
            {'type': 'literal', 'value': Decimal('9')},
            {'type': 'rbrace', 'value': '}'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'comparator',
            'value': 'eq',
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
                {'type': 'literal', 'children': [], 'value': {8, 9}},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_set_literal_can_only_contain_literals(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbrace', 'value': '{'},
            {'type': 'identifier', 'value': 'bar'},
            {'type': 'rbrace', 'value': '}'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_parse_sets_dont_take_floats(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbrace', 'value': '{'},
            {'type': 'literal', 'value': 1.1},
            {'type': 'rbrace', 'value': '}'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(InvalidLiteralValueError):
            self._parse(tokens)

    def test_sets_cant_contain_collections(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbrace', 'value': '{'},
            {'type': 'lbracket', 'value': '['},
            {'type': 'literal', 'value': 9},
            {'type': 'rbracket', 'value': ']'},
            {'type': 'rbrace', 'value': '}'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_set_values_must_all_be_same_type(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbrace', 'value': '{'},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'comma', 'value': ','},
            {'type': 'literal', 'value': Decimal('9')},
            {'type': 'rbrace', 'value': '}'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(InvalidLiteralValueError):
            self._parse(tokens)

    def test_set_with_unmatched_brace(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbrace', 'value': '{'},
            {'type': 'literal', 'value': Decimal('8')},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_parse_map_literal(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbrace', 'value': '{'},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'colon', 'value': ':'},
            {'type': 'literal', 'value': 4},
            {'type': 'comma', 'value': ','},
            {'type': 'literal', 'value': 'baz'},
            {'type': 'colon', 'value': ':'},
            {'type': 'literal', 'value': 3},
            {'type': 'rbrace', 'value': '}'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'comparator',
            'value': 'eq',
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
                {
                    'type': 'literal',
                    'children': [],
                    'value': {"bar": 4, "baz": 3},
                },
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_empty_map(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbrace', 'value': '{'},
            {'type': 'rbrace', 'value': '}'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'comparator',
            'value': 'eq',
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
                {'type': 'literal', 'children': [], 'value': {}},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_map_with_unmatched_brace(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbrace', 'value': '{'},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'colon', 'value': ':'},
            {'type': 'literal', 'value': 4},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_map_missing_colon(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbrace', 'value': '{'},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'literal', 'value': 4},
            {'type': 'rbrace', 'value': '}'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_map_key_must_be_string(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'eq', 'value': '='},
            {'type': 'lbrace', 'value': '{'},
            {'type': 'literal', 'value': 9},
            {'type': 'colon', 'value': ':'},
            {'type': 'literal', 'value': 4},
            {'type': 'rbrace', 'value': '}'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(InvalidLiteralValueError):
            self._parse(tokens)

    def test_parse_empty_expression(self):
        tokens = [{'type': 'eof', 'value': ''}]
        with self.assertRaises(EmptyExpressionError):
            self._parse(tokens)

    def test_eof_expected_but_not_matched(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'comma', 'value': ','},
            {'type': 'identifier', 'value': 'bar'},
            {'type': 'identifier', 'value': 'baz'},
            {'type': 'eof', 'value': ''},
        ]
        expected = 'Expected type: eof'
        with self.assertRaisesRegex(UnexpectedTokenError, expected):
            self._parse(tokens)

    def test_missing_eof(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'comma', 'value': ','},
            {'type': 'identifier', 'value': 'bar'},
        ]
        with self.assertRaises(UnexpectedTokenError):
            self._parse(tokens)

    def test_dotted_identifier(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'dot', 'value': '.'},
            {'type': 'identifier', 'value': 'bar'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'path_identifier',
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
                {'type': 'identifier', 'value': 'bar', 'children': []},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_index_identifier(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'lbracket', 'value': '['},
            {'type': 'literal', 'value': Decimal(0)},
            {'type': 'rbracket', 'value': ']'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'index_identifier',
            'value': Decimal(0),
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_index_identifier_must_be_whole_number(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'lbracket', 'value': '['},
            {'type': 'literal', 'value': Decimal(1.1)},
            {'type': 'rbracket', 'value': ']'},
            {'type': 'eof', 'value': ''},
        ]
        with self.assertRaises(InvalidLiteralValueError):
            self._parse(tokens)

    def test_dotted_index_identifier(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'lbracket', 'value': '['},
            {'type': 'literal', 'value': Decimal(0)},
            {'type': 'rbracket', 'value': ']'},
            {'type': 'dot', 'value': '.'},
            {'type': 'identifier', 'value': 'bar'},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'path_identifier',
            'children': [
                {
                    'type': 'index_identifier',
                    'value': Decimal(0),
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                    ],
                },
                {'type': 'identifier', 'value': 'bar', 'children': []},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_between_complex_identifier(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'dot', 'value': '.'},
            {'type': 'identifier', 'value': 'bar'},
            {'type': 'between', 'value': 'between'},
            {'type': 'literal', 'value': 1},
            {'type': 'and', 'value': 'and'},
            {'type': 'literal', 'value': 3},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'between_expression',
            'children': [
                {
                    'type': 'path_identifier',
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                        {'type': 'identifier', 'value': 'bar', 'children': []},
                    ],
                },
                {'type': 'literal', 'value': 1, 'children': []},
                {'type': 'literal', 'value': 3, 'children': []},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_in_complex_identifier(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'lbracket', 'value': '['},
            {'type': 'literal', 'value': Decimal(0)},
            {'type': 'rbracket', 'value': ']'},
            {'type': 'in', 'value': 'in'},
            {'type': 'lparen', 'value': '('},
            {'type': 'literal', 'value': 'bar'},
            {'type': 'comma', 'value': ','},
            {'type': 'literal', 'value': 'baz'},
            {'type': 'rparen', 'value': ')'},
            {'type': 'eof', 'value': ''},
        ]
        sequence = {
            'type': 'sequence',
            'children': [
                {'type': 'literal', 'value': 'bar', 'children': []},
                {'type': 'literal', 'value': 'baz', 'children': []},
            ],
        }
        expected = {
            'type': 'in_expression',
            'children': [
                {
                    'type': 'index_identifier',
                    'value': Decimal(0),
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                    ],
                },
                sequence,
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_compare_complex_identifier(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'dot', 'value': '.'},
            {'type': 'identifier', 'value': 'bar'},
            {'type': 'gte', 'value': '>='},
            {'type': 'literal', 'value': 8},
            {'type': 'eof', 'value': ''},
        ]
        expected = {
            'type': 'comparator',
            'value': 'gte',
            'children': [
                {
                    'type': 'path_identifier',
                    'children': [
                        {'type': 'identifier', 'value': 'foo', 'children': []},
                        {'type': 'identifier', 'value': 'bar', 'children': []},
                    ],
                },
                {'type': 'literal', 'value': 8, 'children': []},
            ],
        }
        self.assert_parse(tokens, expected)

    def test_parse_complex_identifier_sequence(self):
        tokens = [
            {'type': 'identifier', 'value': 'foo'},
            {'type': 'dot', 'value': '.'},
            {'type': 'identifier', 'value': 'bar'},
            {'type': 'comma', 'value': ','},
            {'type': 'identifier', 'value': 'baz'},
            {'type': 'eof', 'value': ''},
        ]
        first = {
            'type': 'path_identifier',
            'children': [
                {'type': 'identifier', 'value': 'foo', 'children': []},
                {'type': 'identifier', 'value': 'bar', 'children': []},
            ],
        }
        expected = {
            'type': 'sequence',
            'children': [
                first,
                {'type': 'identifier', 'value': 'baz', 'children': []},
            ],
        }
        self.assert_parse(tokens, expected)



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.session import Session

from awscli.paramfile import (
    LOCAL_PREFIX_MAP,
    ResourceLoadingError,
    get_paramfile,
    register_uri_param_handler,
)
from awscli.testutils import FileCreator, mock, skip_if_windows, unittest


class TestParamFile(unittest.TestCase):
    def setUp(self):
        self.files = FileCreator()

    def tearDown(self):
        self.files.remove_all()

    def get_paramfile(self, path):
        return get_paramfile(path, LOCAL_PREFIX_MAP.copy())

    def test_text_file(self):
        contents = 'This is a test'
        filename = self.files.create_file('foo', contents)
        prefixed_filename = 'file://' + filename
        data = self.get_paramfile(prefixed_filename)
        self.assertEqual(data, contents)
        self.assertIsInstance(data, str)

    def test_binary_file(self):
        contents = 'This is a test'
        filename = self.files.create_file('foo', contents)
        prefixed_filename = 'fileb://' + filename
        data = self.get_paramfile(prefixed_filename)
        self.assertEqual(data, b'This is a test')
        self.assertIsInstance(data, bytes)

    @skip_if_windows(
        'Binary content error only occurs ' 'on non-Windows platforms.'
    )
    def test_cannot_load_text_file(self):
        contents = b'\xbfX\xac\xbe'
        filename = self.files.create_file('foo', contents, mode='wb')
        prefixed_filename = 'file://' + filename
        with self.assertRaises(ResourceLoadingError):
            self.get_paramfile(prefixed_filename)

    def test_file_does_not_exist_raises_error(self):
        with self.assertRaises(ResourceLoadingError):
            self.get_paramfile('file://file/does/not/existsasdf.txt')

    def test_no_match_uris_returns_none(self):
        self.assertIsNone(self.get_paramfile('foobar://somewhere.bar'))

    def test_non_string_type_returns_none(self):
        self.assertIsNone(self.get_paramfile(100))


class TestConfigureURIArgumentHandler(unittest.TestCase):
    @mock.patch('awscli.paramfile.URIArgumentHandler')
    def test_default_prefix_maps(self, mock_handler_cls):
        session = mock.Mock(spec=Session)
        session.get_scoped_config.return_value = {}

        register_uri_param_handler(session)
        cases = mock_handler_cls.call_args[0][0]

        self.assertIn('file://', cases)
        self.assertIn('fileb://', cases)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime

from botocore.compat import parse_qs, urlparse
from dateutil.tz import tzutc

from awscli.testutils import BaseAWSCommandParamsTest, mock


class TestGenerateDBAuthToken(BaseAWSCommandParamsTest):
    prefix = 'rds generate-db-auth-token'

    def _urlparse(self, url):
        if isinstance(url, bytes):
            # Not really necessary, but it helps to reduce noise on Python 2.x
            url = url.decode('utf8')
        return urlparse(url)

    def assert_url_equal(self, url1, url2):
        parts1 = self._urlparse(url1)
        parts2 = self._urlparse(url2)

        # Because the query string ordering isn't relevant, we have to parse
        # every single part manually and then handle the query string.
        self.assertEqual(parts1.scheme, parts2.scheme)
        self.assertEqual(parts1.netloc, parts2.netloc)
        self.assertEqual(parts1.path, parts2.path)
        self.assertEqual(parts1.params, parts2.params)
        self.assertEqual(parts1.fragment, parts2.fragment)
        self.assertEqual(parts1.username, parts2.username)
        self.assertEqual(parts1.password, parts2.password)
        self.assertEqual(parts1.hostname, parts2.hostname)
        self.assertEqual(parts1.port, parts2.port)
        self.assertEqual(parse_qs(parts1.query), parse_qs(parts2.query))

    def test_generate_simple_token(self):
        command = self.prefix + ' --hostname host.us-east-1.amazonaws.com'
        command += ' --port 3306 --username mySQLUser'
        clock = datetime.datetime(2016, 11, 7, 17, 39, 33, tzinfo=tzutc())

        with mock.patch('datetime.datetime') as dt:
            dt.utcnow.return_value = clock
            stdout, _, _ = self.run_cmd(command, expected_rc=0)

        expected = (
            'host.us-east-1.amazonaws.com:3306/?DBUser=mySQLUser&'
            'Action=connect&X-Amz-Credential=access_key%2F20161107%2Fus-east-1'
            '%2Frds-db%2Faws4_request&X-Amz-Expires=900&X-Amz-Date=20161107T173'
            '933Z&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-SignedHeaders=host&'
            'X-Amz-Signature=87ab58107ef49f1c311a412f98b7f976b0b5152ffb559f0d'
            '36c6c9a0c5e0e362'
        )

        self.assert_url_equal(
            'https://' + stdout.strip('\n'), 'https://' + expected
        )



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime

from awscli.customizations.s3.filegenerator import FileStat
from awscli.customizations.s3.syncstrategy.sizeonly import SizeOnlySync
from awscli.testutils import unittest


class TestSizeOnlySync(unittest.TestCase):
    def setUp(self):
        self.sync_strategy = SizeOnlySync()

    def test_compare_size_only(self):
        """
        Confirm that files are synced when size differs.
        """
        time_src = datetime.datetime.now()
        time_dst = time_src + datetime.timedelta(days=1)

        src_file = FileStat(
            src='',
            dest='',
            compare_key='test.py',
            size=11,
            last_update=time_src,
            src_type='local',
            dest_type='s3',
            operation_name='upload',
        )

        dst_file = FileStat(
            src='',
            dest='',
            compare_key='test.py',
            size=10,
            last_update=time_dst,
            src_type='s3',
            dest_type='local',
            operation_name='',
        )

        should_sync = self.sync_strategy.determine_should_sync(
            src_file, dst_file
        )
        self.assertTrue(should_sync)

    def test_compare_size_only_different_update_times(self):
        """
        Confirm that files with the same size but different update times
        are not synced.
        """
        time_src = datetime.datetime.now()
        time_dst = time_src + datetime.timedelta(days=1)

        src_file = FileStat(
            src='',
            dest='',
            compare_key='test.py',
            size=10,
            last_update=time_src,
            src_type='local',
            dest_type='s3',
            operation_name='upload',
        )

        dst_file = FileStat(
            src='',
            dest='',
            compare_key='test.py',
            size=10,
            last_update=time_dst,
            src_type='s3',
            dest_type='local',
            operation_name='',
        )

        should_sync = self.sync_strategy.determine_should_sync(
            src_file, dst_file
        )
        self.assertFalse(should_sync)


if __name__ == "__main__":
    unittest.main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest, FileCreator


class BaseQuickSightAssetBundleTest(BaseAWSCommandParamsTest):
    def setUp(self):
        super(BaseQuickSightAssetBundleTest, self).setUp()
        self.files = FileCreator()
        self.temp_file = self.files.create_file('foo', 'mycontents')
        with open(self.temp_file, 'rb') as f:
            self.temp_file_bytes = f.read()

    def tearDown(self):
        super(BaseQuickSightAssetBundleTest, self).tearDown()
        self.files.remove_all()


class TestStartAssetBundleImportJob(BaseQuickSightAssetBundleTest):
    prefix = (
        'quicksight start-asset-bundle-import-job '
        '--aws-account-id 123456789012 '
        '--asset-bundle-import-job-id import-job-1 '
    )

    def test_can_provide_source_body_as_top_level_param(self):
        cmdline = self.prefix
        cmdline += (
            f' --asset-bundle-import-source-bytes fileb://{self.temp_file}'
        )
        result = {
            'AwsAccountId': '123456789012',
            'AssetBundleImportJobId': 'import-job-1',
            'AssetBundleImportSource': {'Body': self.temp_file_bytes},
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_can_provide_source_body_as_nested_param(self):
        cmdline = self.prefix
        cmdline += ' --asset-bundle-import-source Body=aGVsbG8gd29ybGQ='
        result = {
            'AwsAccountId': '123456789012',
            'AssetBundleImportJobId': 'import-job-1',
            'AssetBundleImportSource': {'Body': b'hello world'},
        }
        self.assert_params_for_cmd(cmdline, result)



# Copyright 2012-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations.servicecatalog import utils
from awscli.testutils import unittest


class TestUtils(unittest.TestCase):
    def test_make_url_no_version(self):
        url = utils.make_url('us-east-1', 'foo-bucket-name', 'a/b/c', None)
        self.assertEqual("https://s3.amazonaws.com/foo-bucket-name/a/b/c", url)

    def test_make_url_no_version_other_region(self):
        url = utils.make_url('us-west-2', 'foo-bucket-name', 'a/b/c', None)
        self.assertEqual(
            "https://s3-us-west-2.amazonaws.com/foo-bucket-name/a/b/c", url
        )

    def test_make_url_with_version(self):
        url = utils.make_url('us-east-1', 'foo-bucket-name', 'a/b/c', 5)
        self.assertEqual(
            "https://s3.amazonaws.com/foo-bucket-name/a/b/c?versionId=5", url
        )



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

DEFAULT_INSTANCE_FLEET_NAME = "if-XYZ123"
DEFAULT_CLUSTER_NAME = "j-ABC123456"

INSTANCE_FLEETS_WITH_ON_DEMAND_MASTER_ONLY = (
    'InstanceFleetType=MASTER,TargetOnDemandCapacity=1,InstanceTypeConfigs=[{InstanceType=d2.xlarge}],'
    'LaunchSpecifications={OnDemandSpecification={AllocationStrategy=lowest-price,'
    'CapacityReservationOptions={UsageStrategy=use-capacity-reservations-first,CapacityReservationPreference=open}}},'
    'ResizeSpecifications={OnDemandResizeSpecification={TimeoutDurationMinutes=10}}'
)

INSTANCE_FLEETS_WITH_ON_DEMAND_MASTER_ONLY_WITH_TARGETED_ODCR = (
    'InstanceFleetType=MASTER,TargetOnDemandCapacity=1,InstanceTypeConfigs=[{InstanceType=d2.xlarge}],'
    'LaunchSpecifications={OnDemandSpecification={AllocationStrategy=lowest-price,'
    'CapacityReservationOptions={UsageStrategy=use-capacity-reservations-first,CapacityReservationResourceGroupArn=arn:aws:resource-groups:us-east-1:123456789012:group/Test}}}'
)

INSTANCE_FLEETS_WITH_SPOT_MASTER_ONLY = (
    'InstanceFleetType=MASTER,TargetSpotCapacity=1,InstanceTypeConfigs=[{InstanceType=d2.xlarge,BidPrice=0.1}],'
    'LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=20,TimeoutAction=TERMINATE_CLUSTER,AllocationStrategy=capacity-optimized}},'
    'ResizeSpecifications={SpotResizeSpecification={TimeoutDurationMinutes=10}}'
)

INSTANCE_FLEETS_WITH_SPOT_MASTER_ONLY_WITH_EBS_CONF = (
    'InstanceFleetType=MASTER,TargetSpotCapacity=1,InstanceTypeConfigs=[{InstanceType=d2.xlarge,BidPrice=0.1,'
    'EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={VolumeType=gp2,SizeInGB=100,'
    'Iops=100},VolumesPerInstance=4},{VolumeSpecification={VolumeType=gp2,SizeInGB=100,Iops=100}}]}}],'
    'LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=20,TimeoutAction=TERMINATE_CLUSTER,AllocationStrategy=capacity-optimized}}'
)

INSTANCE_FLEETS_WITH_SPOT_MASTER_CORE_CLUSTER = (
    'InstanceFleetType=MASTER,TargetSpotCapacity=1,InstanceTypeConfigs=[{InstanceType=d2.xlarge,BidPrice=0.1}],'
    'LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=20,TimeoutAction=TERMINATE_CLUSTER}},'
    'ResizeSpecifications={SpotResizeSpecification={TimeoutDurationMinutes=10}} '
    'InstanceFleetType=CORE,TargetSpotCapacity=100,InstanceTypeConfigs=[{InstanceType=d2.xlarge,BidPrice=0.5,'
    'WeightedCapacity=1},{InstanceType=m3.2xlarge,BidPrice=0.2,WeightedCapacity=2},{InstanceType=m3.4xlarge,BidPrice=0.4,'
    'WeightedCapacity=4}],LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=20,'
    'TimeoutAction=SWITCH_TO_ON_DEMAND,'
    'BlockDurationMinutes=120}},'
    'ResizeSpecifications={OnDemandResizeSpecification={TimeoutDurationMinutes=20},SpotResizeSpecification={TimeoutDurationMinutes=30}}'
)

INSTANCE_FLEETS_WITH_SPOT_MASTER_CORE_CLUSTER_WITH_CUSTOM_AMI = (
    'InstanceFleetType=MASTER,TargetSpotCapacity=1,InstanceTypeConfigs=[{InstanceType=d2.xlarge,BidPrice=0.1,CustomAmiId=ami-deadbeef}],'
    'LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=20,TimeoutAction=TERMINATE_CLUSTER}} '
    'InstanceFleetType=CORE,TargetSpotCapacity=100,InstanceTypeConfigs=[{InstanceType=d2.xlarge,BidPrice=0.5,'
    'WeightedCapacity=1,CustomAmiId=ami-deadbeef},{InstanceType=m3.2xlarge,BidPrice=0.2,WeightedCapacity=2,CustomAmiId=ami-deadpork},{InstanceType=m3.4xlarge,BidPrice=0.4,'
    'WeightedCapacity=4,CustomAmiId=ami-deadpork}],LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=20,'
    'TimeoutAction=SWITCH_TO_ON_DEMAND,'
    'BlockDurationMinutes=120}}'
)

INSTANCE_FLEETS_WITH_MISSING_BID_PRICE_CONFIGS = (
    'InstanceFleetType=MASTER,TargetSpotCapacity=1,InstanceTypeConfigs=[{InstanceType=d2.xlarge}],'
    'LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=20,TimeoutAction=TERMINATE_CLUSTER}}'
)

INSTANCE_FLEETS_WITH_SPOT_ALLOCATION_STRATEGY = (
    'InstanceFleetType=MASTER,TargetSpotCapacity=1,InstanceTypeConfigs=[{InstanceType=d2.xlarge,BidPrice=0.1}],'
    'LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=20,TimeoutAction=TERMINATE_CLUSTER,AllocationStrategy=price-capacity-optimized}} '
    'InstanceFleetType=CORE,TargetSpotCapacity=100,InstanceTypeConfigs=[{InstanceType=d2.xlarge,BidPrice=0.5,'
    'WeightedCapacity=1},{InstanceType=m3.2xlarge,BidPrice=0.2,WeightedCapacity=2},{InstanceType=m3.4xlarge,BidPrice=0.4,'
    'WeightedCapacity=4}],LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=20,'
    'TimeoutAction=TERMINATE_CLUSTER,AllocationStrategy=lowest-price}} '
    'InstanceFleetType=TASK,TargetSpotCapacity=100,InstanceTypeConfigs=[{InstanceType=d2.xlarge,BidPrice=0.5,'
    'WeightedCapacity=1},{InstanceType=m3.2xlarge,BidPrice=0.2,WeightedCapacity=2},{InstanceType=m3.4xlarge,BidPrice=0.4,'
    'WeightedCapacity=4}],LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=20,'
    'TimeoutAction=TERMINATE_CLUSTER,AllocationStrategy=diversified}}'
)

INSTANCE_FLEETS_WITH_PRIORITIZED_ALLOCATION_STRATEGY_SPOT_AND_OD = (
    'InstanceFleetType=MASTER,TargetSpotCapacity=1,InstanceTypeConfigs=[{InstanceType=d2.xlarge,'
    'BidPrice=0.1,Priority=0.0}],LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=30,'
    'TimeoutAction=TERMINATE_CLUSTER,AllocationStrategy=capacity-optimized-prioritized},'
    'OnDemandSpecification={AllocationStrategy=prioritized}} '
    'InstanceFleetType=CORE,TargetSpotCapacity=100,InstanceTypeConfigs=[{InstanceType=d2.xlarge,'
    'BidPrice=0.5,WeightedCapacity=1,Priority=0.0},{InstanceType=m3.2xlarge,BidPrice=0.2,'
    'WeightedCapacity=2,Priority=1.0},{InstanceType=m3.4xlarge,BidPrice=0.4,WeightedCapacity=4,'
    'Priority=99.0}],LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=32,'
    'TimeoutAction=TERMINATE_CLUSTER,AllocationStrategy=capacity-optimized-prioritized},'
    'OnDemandSpecification={AllocationStrategy=prioritized}} '
    'InstanceFleetType=TASK,TargetSpotCapacity=100,InstanceTypeConfigs=[{InstanceType=d2.xlarge,'
    'BidPrice=0.5,WeightedCapacity=1,Priority=10.0},{InstanceType=m3.2xlarge,BidPrice=0.2,'
    'WeightedCapacity=2,Priority=0.0},{InstanceType=m3.4xlarge,BidPrice=0.4,WeightedCapacity=4,'
    'Priority=100.0}],LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=77,'
    'TimeoutAction=TERMINATE_CLUSTER,AllocationStrategy='
    'capacity-optimized-prioritized},OnDemandSpecification={AllocationStrategy=prioritized}}'
)

TASK_INSTANCE_FLEET_WITH_RESIZE_ALLOCATION_STRATEGY_SPOT_AND_OD = (
    'InstanceFleetType=TASK,TargetSpotCapacity=100,Context=testContext,InstanceTypeConfigs=[{InstanceType=d2.xlarge,'
    'BidPrice=0.5,WeightedCapacity=1},{InstanceType=m3.2xlarge,BidPrice=0.2,WeightedCapacity=2},'
    '{InstanceType=m3.4xlarge,BidPrice=0.4,WeightedCapacity=4}],LaunchSpecifications={'
    'SpotSpecification={TimeoutDurationMinutes=77,TimeoutAction=TERMINATE_CLUSTER,'
    'AllocationStrategy=capacity-optimized-prioritized},OnDemandSpecification={'
    'AllocationStrategy=lowest-price}},ResizeSpecifications={SpotResizeSpecification={'
    'AllocationStrategy=capacity-optimized},OnDemandResizeSpecification={'
    'AllocationStrategy=lowest-price,CapacityReservationOptions={'
    'UsageStrategy=use-capacity-reservations-first,CapacityReservationPreference=open}}}'
)

INSTANCE_FLEETS_WITH_RESIZE_ALLOCATION_STRATEGY_SPOT_AND_OD = (
    'InstanceFleetType=MASTER,TargetSpotCapacity=1,InstanceTypeConfigs=[{InstanceType=d2.xlarge,'
    'BidPrice=0.1}],LaunchSpecifications={SpotSpecification={TimeoutDurationMinutes=30,'
    'TimeoutAction=TERMINATE_CLUSTER,AllocationStrategy=capacity-optimized-prioritized},'
    'OnDemandSpecification={AllocationStrategy=lowest-price}} '
    'InstanceFleetType=CORE,TargetSpotCapacity=100,InstanceTypeConfigs=[{InstanceType=d2.xlarge,'
    'BidPrice=0.5,WeightedCapacity=1},{InstanceType=m3.2xlarge,BidPrice=0.2,WeightedCapacity=2},'
    '{InstanceType=m3.4xlarge,BidPrice=0.4,WeightedCapacity=4}],LaunchSpecifications={'
    'SpotSpecification={TimeoutDurationMinutes=32,TimeoutAction=TERMINATE_CLUSTER,'
    'AllocationStrategy=capacity-optimized-prioritized},OnDemandSpecification={'
    'AllocationStrategy=lowest-price}},ResizeSpecifications={SpotResizeSpecification='
    '{AllocationStrategy=capacity-optimized},OnDemandResizeSpecification={'
    'AllocationStrategy=lowest-price,CapacityReservationOptions={'
    'UsageStrategy=use-capacity-reservations-first,CapacityReservationPreference=open}}} '
    f'{TASK_INSTANCE_FLEET_WITH_RESIZE_ALLOCATION_STRATEGY_SPOT_AND_OD}'
)

MODIFY_INSTANCE_FLEET_WITH_INSTANCE_TYPE_CONFIGS = (
    f'InstanceFleetId={DEFAULT_INSTANCE_FLEET_NAME},'
    f'InstanceTypeConfigs=[{{InstanceType=d2.xlarge}}],Context=testContext'
)

MODIFY_INSTANCE_FLEET_WITH_SPOT_AND_OD_RESIZE_SPECIFICATIONS = (
    f'InstanceFleetId={DEFAULT_INSTANCE_FLEET_NAME},ResizeSpecifications={{SpotResizeSpecification='
    f'{{AllocationStrategy=capacity-optimized}},OnDemandResizeSpecification={{'
    f'AllocationStrategy=lowest-price,CapacityReservationOptions={{'
    f'UsageStrategy=use-capacity-reservations-first,CapacityReservationPreference=open}}}}}}'
)

MODIFY_INSTANCE_FLEET_WITH_INSTANCE_TYPE_CONFIGS_AND_SPOT_AND_OD_RESIZE_SPECIFICATIONS = (
    f'InstanceFleetId={DEFAULT_INSTANCE_FLEET_NAME},ResizeSpecifications={{SpotResizeSpecification='
    f'{{AllocationStrategy=capacity-optimized}},OnDemandResizeSpecification={{'
    f'AllocationStrategy=lowest-price,CapacityReservationOptions={{'
    f'UsageStrategy=use-capacity-reservations-first,CapacityReservationPreference=open}}}}}}'
    f',InstanceTypeConfigs=[{{InstanceType=d2.xlarge}}]'
)

RES_INSTANCE_FLEETS_WITH_ON_DEMAND_MASTER_ONLY = [
    {
        "InstanceTypeConfigs": [{"InstanceType": "d2.xlarge"}],
        "LaunchSpecifications": {
            "OnDemandSpecification": {
                "AllocationStrategy": "lowest-price",
                "CapacityReservationOptions": {
                    "UsageStrategy": "use-capacity-reservations-first",
                    "CapacityReservationPreference": "open",
                },
            }
        },
        "ResizeSpecifications": {
            "OnDemandResizeSpecification": {"TimeoutDurationMinutes": 10}
        },
        "TargetOnDemandCapacity": 1,
        "InstanceFleetType": "MASTER",
        "Name": "MASTER",
    }
]

RES_INSTANCE_FLEETS_WITH_ON_DEMAND_MASTER_ONLY_WITH_TARGETED_ODCR = [
    {
        "InstanceTypeConfigs": [{"InstanceType": "d2.xlarge"}],
        "LaunchSpecifications": {
            "OnDemandSpecification": {
                "AllocationStrategy": "lowest-price",
                "CapacityReservationOptions": {
                    "UsageStrategy": "use-capacity-reservations-first",
                    "CapacityReservationResourceGroupArn": "arn:aws:resource-groups:us-east-1:123456789012:group/Test",
                },
            }
        },
        "TargetOnDemandCapacity": 1,
        "InstanceFleetType": "MASTER",
        "Name": "MASTER",
    }
]

RES_INSTANCE_FLEETS_WITH_SPOT_MASTER_ONLY = [
    {
        "InstanceTypeConfigs": [
            {"InstanceType": "d2.xlarge", "BidPrice": "0.1"}
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 20,
                "TimeoutAction": "TERMINATE_CLUSTER",
                "AllocationStrategy": "capacity-optimized",
            }
        },
        "ResizeSpecifications": {
            "SpotResizeSpecification": {"TimeoutDurationMinutes": 10}
        },
        "TargetSpotCapacity": 1,
        "InstanceFleetType": "MASTER",
        "Name": "MASTER",
    }
]

RES_INSTANCE_FLEETS_WITH_SPOT_MASTER_ONLY_WITH_EBS_CONF = [
    {
        "InstanceTypeConfigs": [
            {
                "InstanceType": "d2.xlarge",
                "BidPrice": "0.1",
                "EbsConfiguration": {
                    "EbsOptimized": True,
                    "EbsBlockDeviceConfigs": [
                        {
                            "VolumeSpecification": {
                                "Iops": 100,
                                "SizeInGB": 100,
                                "VolumeType": "gp2",
                            },
                            "VolumesPerInstance": 4,
                        },
                        {
                            "VolumeSpecification": {
                                "Iops": 100,
                                "SizeInGB": 100,
                                "VolumeType": "gp2",
                            }
                        },
                    ],
                },
            }
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 20,
                "TimeoutAction": "TERMINATE_CLUSTER",
                "AllocationStrategy": "capacity-optimized",
            }
        },
        "TargetSpotCapacity": 1,
        "InstanceFleetType": "MASTER",
        "Name": "MASTER",
    }
]

RES_INSTANCE_FLEETS_WITH_SPOT_MASTER_CORE_CLUSTER = [
    {
        "InstanceTypeConfigs": [
            {"InstanceType": "d2.xlarge", "BidPrice": "0.1"}
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 20,
                "TimeoutAction": "TERMINATE_CLUSTER",
            }
        },
        "ResizeSpecifications": {
            "SpotResizeSpecification": {"TimeoutDurationMinutes": 10}
        },
        "TargetSpotCapacity": 1,
        "InstanceFleetType": "MASTER",
        "Name": "MASTER",
    },
    {
        "InstanceTypeConfigs": [
            {
                "InstanceType": "d2.xlarge",
                "BidPrice": "0.5",
                "WeightedCapacity": 1,
            },
            {
                "InstanceType": "m3.2xlarge",
                "BidPrice": "0.2",
                "WeightedCapacity": 2,
            },
            {
                "InstanceType": "m3.4xlarge",
                "BidPrice": "0.4",
                "WeightedCapacity": 4,
            },
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 20,
                "TimeoutAction": "SWITCH_TO_ON_DEMAND",
                "BlockDurationMinutes": 120,
            }
        },
        "ResizeSpecifications": {
            "OnDemandResizeSpecification": {"TimeoutDurationMinutes": 20},
            "SpotResizeSpecification": {"TimeoutDurationMinutes": 30},
        },
        "TargetSpotCapacity": 100,
        "InstanceFleetType": "CORE",
        "Name": "CORE",
    },
]

RES_INSTANCE_FLEETS_WITH_SPOT_MASTER_CORE_CLUSTER_WITH_CUSTOM_AMI = [
    {
        "InstanceTypeConfigs": [
            {
                "InstanceType": "d2.xlarge",
                "BidPrice": "0.1",
                "CustomAmiId": "ami-deadbeef",
            }
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 20,
                "TimeoutAction": "TERMINATE_CLUSTER",
            }
        },
        "TargetSpotCapacity": 1,
        "InstanceFleetType": "MASTER",
        "Name": "MASTER",
    },
    {
        "InstanceTypeConfigs": [
            {
                "InstanceType": "d2.xlarge",
                "BidPrice": "0.5",
                "WeightedCapacity": 1,
                "CustomAmiId": "ami-deadbeef",
            },
            {
                "InstanceType": "m3.2xlarge",
                "BidPrice": "0.2",
                "WeightedCapacity": 2,
                "CustomAmiId": "ami-deadpork",
            },
            {
                "InstanceType": "m3.4xlarge",
                "BidPrice": "0.4",
                "WeightedCapacity": 4,
                "CustomAmiId": "ami-deadpork",
            },
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 20,
                "TimeoutAction": "SWITCH_TO_ON_DEMAND",
                "BlockDurationMinutes": 120,
            }
        },
        "TargetSpotCapacity": 100,
        "InstanceFleetType": "CORE",
        "Name": "CORE",
    },
]

RES_INSTANCE_FLEETS_WITH_COMPLEX_CONFIG_FROM_JSON = [
    {
        "InstanceTypeConfigs": [
            {
                "EbsConfiguration": {
                    "EbsOptimized": True,
                    "EbsBlockDeviceConfigs": [
                        {
                            "VolumeSpecification": {
                                "VolumeType": "gp2",
                                "SizeInGB": 100,
                            },
                            "VolumesPerInstance": 2,
                        }
                    ],
                },
                "BidPrice": "1",
                "InstanceType": "m3.xlarge",
            }
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 20,
                "TimeoutAction": "TERMINATE_CLUSTER",
            }
        },
        "Name": "master-fleet",
        "InstanceFleetType": "MASTER",
        "TargetSpotCapacity": 1,
    },
    {
        "InstanceTypeConfigs": [
            {
                "WeightedCapacity": 1,
                "EbsConfiguration": {
                    "EbsOptimized": True,
                    "EbsBlockDeviceConfigs": [
                        {
                            "VolumeSpecification": {
                                "VolumeType": "gp2",
                                "SizeInGB": 100,
                            },
                            "VolumesPerInstance": 2,
                        }
                    ],
                },
                "BidPrice": "1",
                "InstanceType": "m3.xlarge",
            },
            {
                "WeightedCapacity": 2,
                "EbsConfiguration": {
                    "EbsOptimized": False,
                    "EbsBlockDeviceConfigs": [
                        {
                            "VolumeSpecification": {
                                "VolumeType": "gp2",
                                "SizeInGB": 100,
                            },
                            "VolumesPerInstance": 2,
                        },
                        {
                            "VolumeSpecification": {
                                "VolumeType": "gp2",
                                "SizeInGB": 500,
                            },
                            "VolumesPerInstance": 1,
                        },
                    ],
                },
                "BidPrice": "1",
                "InstanceType": "m3.large",
            },
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 20,
                "TimeoutAction": "TERMINATE_CLUSTER",
            }
        },
        "Name": "core-fleet",
        "InstanceFleetType": "CORE",
        "TargetSpotCapacity": 10,
    },
]

RES_INSTANCE_FLEETS_WITH_SPOT_ALLOCATION_STRATEGY = [
    {
        "InstanceTypeConfigs": [
            {"InstanceType": "d2.xlarge", "BidPrice": "0.1"}
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 20,
                "TimeoutAction": "TERMINATE_CLUSTER",
                "AllocationStrategy": "price-capacity-optimized",
            }
        },
        "TargetSpotCapacity": 1,
        "InstanceFleetType": "MASTER",
        "Name": "MASTER",
    },
    {
        "InstanceTypeConfigs": [
            {
                "InstanceType": "d2.xlarge",
                "BidPrice": "0.5",
                "WeightedCapacity": 1,
            },
            {
                "InstanceType": "m3.2xlarge",
                "BidPrice": "0.2",
                "WeightedCapacity": 2,
            },
            {
                "InstanceType": "m3.4xlarge",
                "BidPrice": "0.4",
                "WeightedCapacity": 4,
            },
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 20,
                "TimeoutAction": "TERMINATE_CLUSTER",
                "AllocationStrategy": "lowest-price",
            }
        },
        "TargetSpotCapacity": 100,
        "InstanceFleetType": "CORE",
        "Name": "CORE",
    },
    {
        "InstanceTypeConfigs": [
            {
                "InstanceType": "d2.xlarge",
                "BidPrice": "0.5",
                "WeightedCapacity": 1,
            },
            {
                "InstanceType": "m3.2xlarge",
                "BidPrice": "0.2",
                "WeightedCapacity": 2,
            },
            {
                "InstanceType": "m3.4xlarge",
                "BidPrice": "0.4",
                "WeightedCapacity": 4,
            },
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 20,
                "TimeoutAction": "TERMINATE_CLUSTER",
                "AllocationStrategy": "diversified",
            }
        },
        "TargetSpotCapacity": 100,
        "InstanceFleetType": "TASK",
        "Name": "TASK",
    },
]

RES_INSTANCE_FLEETS_WITH_PRIORITIZED_ALLOCATION_STRATEGY_SPOT_AND_OD = [
    {
        "InstanceTypeConfigs": [
            {"InstanceType": "d2.xlarge", "BidPrice": "0.1", "Priority": 0}
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 30,
                "TimeoutAction": "TERMINATE_CLUSTER",
                "AllocationStrategy": "capacity-optimized-prioritized",
            },
            "OnDemandSpecification": {"AllocationStrategy": "prioritized"},
        },
        "TargetSpotCapacity": 1,
        "InstanceFleetType": "MASTER",
        "Name": "MASTER",
    },
    {
        "InstanceTypeConfigs": [
            {
                "InstanceType": "d2.xlarge",
                "BidPrice": "0.5",
                "WeightedCapacity": 1,
                "Priority": 0,
            },
            {
                "InstanceType": "m3.2xlarge",
                "BidPrice": "0.2",
                "WeightedCapacity": 2,
                "Priority": 1,
            },
            {
                "InstanceType": "m3.4xlarge",
                "BidPrice": "0.4",
                "WeightedCapacity": 4,
                "Priority": 99,
            },
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 32,
                "TimeoutAction": "TERMINATE_CLUSTER",
                "AllocationStrategy": "capacity-optimized-prioritized",
            },
            "OnDemandSpecification": {"AllocationStrategy": "prioritized"},
        },
        "TargetSpotCapacity": 100,
        "InstanceFleetType": "CORE",
        "Name": "CORE",
    },
    {
        "InstanceTypeConfigs": [
            {
                "InstanceType": "d2.xlarge",
                "BidPrice": "0.5",
                "WeightedCapacity": 1,
                "Priority": 10,
            },
            {
                "InstanceType": "m3.2xlarge",
                "BidPrice": "0.2",
                "WeightedCapacity": 2,
                "Priority": 0,
            },
            {
                "InstanceType": "m3.4xlarge",
                "BidPrice": "0.4",
                "WeightedCapacity": 4,
                "Priority": 100,
            },
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 77,
                "TimeoutAction": "TERMINATE_CLUSTER",
                "AllocationStrategy": "capacity-optimized-prioritized",
            },
            "OnDemandSpecification": {"AllocationStrategy": "prioritized"},
        },
        "TargetSpotCapacity": 100,
        "InstanceFleetType": "TASK",
        "Name": "TASK",
    },
]

RES_INSTANCE_FLEETS_WITH_RESIZE_ALLOCATION_STRATEGY_SPOT_AND_OD = [
    {
        "InstanceTypeConfigs": [
            {"InstanceType": "d2.xlarge", "BidPrice": "0.1"}
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 30,
                "TimeoutAction": "TERMINATE_CLUSTER",
                "AllocationStrategy": "capacity-optimized-prioritized",
            },
            "OnDemandSpecification": {"AllocationStrategy": "lowest-price"},
        },
        "TargetSpotCapacity": 1,
        "InstanceFleetType": "MASTER",
        "Name": "MASTER",
    },
    {
        "InstanceTypeConfigs": [
            {
                "InstanceType": "d2.xlarge",
                "BidPrice": "0.5",
                "WeightedCapacity": 1,
            },
            {
                "InstanceType": "m3.2xlarge",
                "BidPrice": "0.2",
                "WeightedCapacity": 2,
            },
            {
                "InstanceType": "m3.4xlarge",
                "BidPrice": "0.4",
                "WeightedCapacity": 4,
            },
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 32,
                "TimeoutAction": "TERMINATE_CLUSTER",
                "AllocationStrategy": "capacity-optimized-prioritized",
            },
            "OnDemandSpecification": {"AllocationStrategy": "lowest-price"},
        },
        "ResizeSpecifications": {
            "OnDemandResizeSpecification": {
                "AllocationStrategy": "lowest-price",
                "CapacityReservationOptions": {
                    "CapacityReservationPreference": "open",
                    "UsageStrategy": "use-capacity-reservations-first",
                },
            },
            "SpotResizeSpecification": {
                "AllocationStrategy": "capacity-optimized"
            },
        },
        "TargetSpotCapacity": 100,
        "InstanceFleetType": "CORE",
        "Name": "CORE",
    },
    {
        "InstanceTypeConfigs": [
            {
                "InstanceType": "d2.xlarge",
                "BidPrice": "0.5",
                "WeightedCapacity": 1,
            },
            {
                "InstanceType": "m3.2xlarge",
                "BidPrice": "0.2",
                "WeightedCapacity": 2,
            },
            {
                "InstanceType": "m3.4xlarge",
                "BidPrice": "0.4",
                "WeightedCapacity": 4,
            },
        ],
        "LaunchSpecifications": {
            "SpotSpecification": {
                "TimeoutDurationMinutes": 77,
                "TimeoutAction": "TERMINATE_CLUSTER",
                "AllocationStrategy": "capacity-optimized-prioritized",
            },
            "OnDemandSpecification": {"AllocationStrategy": "lowest-price"},
        },
        "ResizeSpecifications": {
            "OnDemandResizeSpecification": {
                "AllocationStrategy": "lowest-price",
                "CapacityReservationOptions": {
                    "CapacityReservationPreference": "open",
                    "UsageStrategy": "use-capacity-reservations-first",
                },
            },
            "SpotResizeSpecification": {
                "AllocationStrategy": "capacity-optimized"
            },
        },
        "TargetSpotCapacity": 100,
        "InstanceFleetType": "TASK",
        "Context": "testContext",
        "Name": "TASK",
    },
]

RES_TASK_INSTANCE_FLEET_WITH_RESIZE_ALLOCATION_STRATEGY_SPOT_AND_OD = {
    "InstanceTypeConfigs": [
        {
            "InstanceType": "d2.xlarge",
            "BidPrice": "0.5",
            "WeightedCapacity": 1,
        },
        {
            "InstanceType": "m3.2xlarge",
            "BidPrice": "0.2",
            "WeightedCapacity": 2,
        },
        {
            "InstanceType": "m3.4xlarge",
            "BidPrice": "0.4",
            "WeightedCapacity": 4,
        },
    ],
    "LaunchSpecifications": {
        "SpotSpecification": {
            "TimeoutDurationMinutes": 77,
            "TimeoutAction": "TERMINATE_CLUSTER",
            "AllocationStrategy": "capacity-optimized-prioritized",
        },
        "OnDemandSpecification": {"AllocationStrategy": "lowest-price"},
    },
    "ResizeSpecifications": {
        "OnDemandResizeSpecification": {
            "AllocationStrategy": "lowest-price",
            "CapacityReservationOptions": {
                "CapacityReservationPreference": "open",
                "UsageStrategy": "use-capacity-reservations-first",
            },
        },
        "SpotResizeSpecification": {
            "AllocationStrategy": "capacity-optimized"
        },
    },
    "TargetSpotCapacity": 100,
    "InstanceFleetType": "TASK",
    "Context": "testContext",
}

RES_MODIFY_INSTANCE_FLEET_WITH_INSTANCE_TYPE_CONFIGS = {
    "ClusterId": DEFAULT_CLUSTER_NAME,
    "InstanceFleet": {
        "InstanceFleetId": DEFAULT_INSTANCE_FLEET_NAME,
        "InstanceTypeConfigs": [{"InstanceType": "d2.xlarge"}],
        "Context": "testContext",
    },
}

RES_MODIFY_INSTANCE_FLEET_WITH_SPOT_AND_OD_RESIZE_SPECIFICATIONS = {
    "ClusterId": DEFAULT_CLUSTER_NAME,
    "InstanceFleet": {
        "InstanceFleetId": DEFAULT_INSTANCE_FLEET_NAME,
        "ResizeSpecifications": {
            "OnDemandResizeSpecification": {
                "AllocationStrategy": "lowest-price",
                "CapacityReservationOptions": {
                    "CapacityReservationPreference": "open",
                    "UsageStrategy": "use-capacity-reservations-first",
                },
            },
            "SpotResizeSpecification": {
                "AllocationStrategy": "capacity-optimized"
            },
        },
    },
}

RES_MODIFY_INSTANCE_FLEET_WITH_INSTANCE_TYPE_CONFIGS_AND_SPOT_AND_OD_RESIZE_SPECIFICATIONS = {
    "ClusterId": DEFAULT_CLUSTER_NAME,
    "InstanceFleet": {
        "InstanceFleetId": DEFAULT_INSTANCE_FLEET_NAME,
        "ResizeSpecifications": {
            "OnDemandResizeSpecification": {
                "AllocationStrategy": "lowest-price",
                "CapacityReservationOptions": {
                    "CapacityReservationPreference": "open",
                    "UsageStrategy": "use-capacity-reservations-first",
                },
            },
            "SpotResizeSpecification": {
                "AllocationStrategy": "capacity-optimized"
            },
        },
        "InstanceTypeConfigs": [{"InstanceType": "d2.xlarge"}],
    },
}



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os
import shutil
import tempfile

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion import Completion
from prompt_toolkit.document import Document
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.history import History

from awscli.autoprompt.history import HistoryCompleter, HistoryDriver
from awscli.testutils import mock, skip_if_windows, unittest


class TestHistoryCompleter(unittest.TestCase):
    def setUp(self):
        self.strings = [
            'chime associate-phone-number-with-user',
            'deploy add-tags-to-on-premises-instances',
            's3 ls',
            'ecs get-task-protection',
            'ecr get-lifecycle-policy-preview',
        ]
        self.buffer = mock.Mock(spec=Buffer)
        history = mock.Mock(spec=History)
        history.get_strings.return_value = self.strings
        self.buffer.history = history
        self.completer = HistoryCompleter(buffer=self.buffer)

    def test_get_completions_on_empty_document(self):
        document = Document('')
        strings = self.completer.get_completions(document)
        self.assertEqual(
            list(strings),
            [
                Completion(
                    text='ecr get-lifecycle-policy-preview',
                    start_position=0,
                    display=FormattedText(
                        [('', 'ecr get-lifecycle-policy-preview')]
                    ),
                ),
                Completion(
                    text='ecs get-task-protection',
                    start_position=0,
                    display=FormattedText([('', 'ecs get-task-protection')]),
                ),
                Completion(
                    text='s3 ls',
                    start_position=0,
                    display=FormattedText([('', 's3 ls')]),
                ),
                Completion(
                    text='deploy add-tags-to-on-premises-instances',
                    start_position=0,
                    display=FormattedText(
                        [('', 'deploy add-tags-to-on-premises-instances')]
                    ),
                ),
                Completion(
                    text='chime associate-phone-number-with-user',
                    start_position=0,
                    display=FormattedText(
                        [('', 'chime associate-phone-number-with-user')]
                    ),
                ),
            ],
        )

    def test_get_completions_with_fuzzy_search(self):
        document = Document('ecgpr')
        strings = self.completer.get_completions(document)
        self.assertEqual(
            list(strings),
            [
                Completion(
                    text='ecs get-task-protection',
                    start_position=-5,
                    display=FormattedText([('', 'ecs get-task-protection')]),
                ),
                Completion(
                    text='ecr get-lifecycle-policy-preview',
                    start_position=-5,
                    display=FormattedText(
                        [('', 'ecr get-lifecycle-policy-preview')]
                    ),
                ),
            ],
        )

    def test_show_newest_duplicated_command(self):
        strings = [
            'chime associate-phone-number-with-user',
            's3 ls',
            's3api list-buckets',
            's3 ls',
        ]
        history = mock.Mock(spec=History)
        history.get_strings.return_value = strings
        self.buffer.history = history
        self.completer = HistoryCompleter(buffer=self.buffer)
        document = Document('')
        response = self.completer.get_completions(document)
        self.assertEqual(
            list(response),
            [
                Completion(
                    text='s3 ls',
                    start_position=0,
                    display=FormattedText([('', 's3 ls')]),
                ),
                Completion(
                    text='s3api list-buckets',
                    start_position=0,
                    display=FormattedText([('', 's3api list-buckets')]),
                ),
                Completion(
                    text='chime associate-phone-number-with-user',
                    start_position=0,
                    display=FormattedText(
                        [('', 'chime associate-phone-number-with-user')]
                    ),
                ),
            ],
        )


class TestHistoryDriver(unittest.TestCase):
    def setUp(self):
        self.dirname = tempfile.mkdtemp()
        self.filename = os.path.join(self.dirname, 'prompt_history.json')
        self.history_driver = HistoryDriver(self.filename)

    def tearDown(self):
        shutil.rmtree(self.dirname)

    def test_store_string(self):
        self.history_driver.store_string('aws ec2 describe-instances')
        with open(self.filename) as f:
            history = json.load(f)
        self.assertEqual(
            history, {'version': 1, 'commands': ['aws ec2 describe-instances']}
        )

    def test_creates_folder_and_file(self):
        non_existing_path = os.path.join(self.dirname, 'foo', 'history.json')
        history_driver = HistoryDriver(non_existing_path)
        history_driver.store_string('aws ec2 describe-instances')
        self.assertTrue(os.path.exists(non_existing_path))

    def test_read_after_write_in_reversed_order(self):
        history_driver = HistoryDriver(self.filename)
        history_driver.store_string('aws ec2 describe-instances')
        history_driver.store_string('aws s3 ls')
        history_driver.store_string('aws dynamodb create-table')
        commands = history_driver.load_history_strings()
        self.assertEqual(
            list(commands),
            [
                'aws dynamodb create-table',
                'aws s3 ls',
                'aws ec2 describe-instances',
            ],
        )

    def test_keep_last_max_commands(self):
        history_driver = HistoryDriver(self.filename, max_commands=2)
        history_driver.store_string('aws ec2 describe-instances')
        history_driver.store_string('aws s3 ls')
        history_driver.store_string('aws dynamodb create-table')
        with open(self.filename) as f:
            commands = json.load(f)['commands']
        self.assertEqual(commands, ['aws s3 ls', 'aws dynamodb create-table'])

    @mock.patch('awscli.autoprompt.history.FileHistory')
    def test_handle_io_errors(self, file_history_mock):
        history_driver = HistoryDriver(self.filename)
        file_history_mock.store_string.side_effect = IOError
        history_driver.store_string('aws dynamodb create-table')

    @skip_if_windows('chmod is not supported on Windows')
    def test_sets_file_permissions_to_0600(self):
        self.history_driver.store_string('aws ec2 describe-instances')
        permissions = os.stat(self.filename).st_mode & 0o777
        self.assertEqual(permissions, 0o600)



import platform

import pytest

skip_if_windows = pytest.mark.skipif(
    platform.system() not in ['Darwin', 'Linux'],
    reason="This test does not run on windows.",
)
if_windows = pytest.mark.skipif(
    platform.system() in ['Darwin', 'Linux'],
    reason="This test only runs on windows.",
)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from botocore.model import OperationModel, ServiceModel, StringShape

from awscli import arguments
from awscli.testutils import mock, unittest


class DemoArgument(arguments.CustomArgument):
    pass


class TestArgumentClasses(unittest.TestCase):
    def test_can_set_required(self):
        arg = DemoArgument('test-arg')
        self.assertFalse(arg.required)
        arg.required = True
        self.assertTrue(arg.required)


class TestCLIArgument(unittest.TestCase):
    def setUp(self):
        self.service_name = 'baz'
        self.service_model = ServiceModel(
            {
                'metadata': {
                    'endpointPrefix': 'bad',
                },
                'operations': {
                    'SampleOperation': {
                        'name': 'SampleOperation',
                        'input': {'shape': 'Input'},
                    }
                },
                'shapes': {
                    'StringShape': {'type': 'string'},
                    'Input': {
                        'type': 'structure',
                        'members': {'Foo': {'shape': 'StringShape'}},
                    },
                },
            },
            self.service_name,
        )
        self.operation_model = self.service_model.operation_model(
            'SampleOperation'
        )
        self.argument_model = self.operation_model.input_shape.members['Foo']
        self.event_emitter = mock.Mock()

    def create_argument(self):
        return arguments.CLIArgument(
            self.argument_model.name,
            self.argument_model,
            self.operation_model,
            self.event_emitter,
        )

    def test_unpack_uses_service_name_in_event(self):
        self.event_emitter.emit.return_value = ['value']
        argument = self.create_argument()
        params = {}
        argument.add_to_params(params, 'value')
        expected_event_name = 'process-cli-arg.%s.%s' % (
            self.service_name,
            'sample-operation',
        )
        actual_event_name = self.event_emitter.emit.call_args[0][0]
        self.assertEqual(actual_event_name, expected_event_name)

    def test_list_type_has_correct_nargs_value(self):
        # We don't actually care about the values, we just need a ListArgument
        # type.
        arg = arguments.ListArgument(
            argument_model=self.argument_model,
            event_emitter=self.event_emitter,
            is_required=True,
            name='test-nargs',
            operation_model=None,
            serialized_name='TestNargs',
        )
        self.assertEqual(arg.nargs, '*')



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import subprocess

from botocore.session import Session

from awscli.alias import (
    AliasCommandInjector,
    AliasLoader,
    AliasSubCommandInjector,
    BaseAliasCommand,
    ExternalAliasCommand,
    InternalAliasSubCommand,
    InvalidAliasException,
    ServiceAliasCommand,
)
from awscli.argparser import ArgParseException, MainArgParser
from awscli.clidriver import CLIDriver
from awscli.commands import CLICommand
from awscli.testutils import FileCreator, mock, unittest


class FakeParsedArgs:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.__dict__)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __contains__(self, key):
        return key in self.__dict__


class TestAliasLoader(unittest.TestCase):
    def setUp(self):
        self.files = FileCreator()
        self.alias_file = self.files.create_file('alias', '[toplevel]\n')

    def tearDown(self):
        self.files.remove_all()

    def test_get_aliases_non_existent_file(self):
        nonexistent_file = os.path.join(self.files.rootdir, 'no-exists')
        alias_interface = AliasLoader(nonexistent_file)
        self.assertEqual(alias_interface.get_aliases(), {})

    def test_get_aliases_empty_file(self):
        alias_interface = AliasLoader(self.alias_file)
        self.assertEqual(alias_interface.get_aliases(), {})

    def test_get_aliases_missing_toplevel(self):
        with open(self.alias_file, 'w') as f:
            f.write('[unrelated-section]\n')
        alias_interface = AliasLoader(self.alias_file)
        self.assertEqual(alias_interface.get_aliases(), {})

    def test_get_aliases(self):
        with open(self.alias_file, 'a+') as f:
            f.write('my-alias = my-alias-value')
        alias_interface = AliasLoader(self.alias_file)
        self.assertEqual(
            alias_interface.get_aliases(), {'my-alias': 'my-alias-value'}
        )

    def test_get_aliases_with_alias_that_includes_parameter(self):
        with open(self.alias_file, 'a+') as f:
            f.write('my-alias = my-alias-value --my-parameter my-value')
        alias_interface = AliasLoader(self.alias_file)
        self.assertEqual(
            alias_interface.get_aliases(),
            {'my-alias': 'my-alias-value --my-parameter my-value'},
        )

    def test_get_aliases_with_alias_that_includes_newlines(self):
        with open(self.alias_file, 'a+') as f:
            f.write('my-alias = my-alias-value\n')
        alias_interface = AliasLoader(self.alias_file)
        self.assertEqual(
            alias_interface.get_aliases(), {'my-alias': 'my-alias-value'}
        )

    def test_get_aliases_with_alias_that_is_indented(self):
        with open(self.alias_file, 'a+') as f:
            f.write('my-alias = \n  my-alias-value\n')
        alias_interface = AliasLoader(self.alias_file)
        self.assertEqual(
            alias_interface.get_aliases(), {'my-alias': 'my-alias-value'}
        )

    def test_get_aliases_with_multiple_lines(self):
        with open(self.alias_file, 'a+') as f:
            f.write(
                'my-alias = \n' '  my-alias-value \\ \n' '  --parameter foo\n'
            )
        alias_interface = AliasLoader(self.alias_file)
        self.assertEqual(
            alias_interface.get_aliases(),
            # The backslash and newline should be preserved, but
            # still have the beginning and end newlines removed.
            {'my-alias': 'my-alias-value \\\n--parameter foo'},
        )

    def test_get_aliases_with_multiple_aliases(self):
        with open(self.alias_file, 'a+') as f:
            f.write('my-alias = my-alias-value\n')
            f.write('my-second-alias = my-second-alias-value\n')
        alias_interface = AliasLoader(self.alias_file)
        self.assertEqual(
            alias_interface.get_aliases(),
            {
                'my-alias': 'my-alias-value',
                'my-second-alias': 'my-second-alias-value',
            },
        )

    def test_get_aliases_for_service(self):
        with open(self.alias_file, 'a+') as f:
            f.write('[command ec2]\n')
            f.write('list = describe-instances\n')
            f.write('regions = describe-regions\n')
        alias_interface = AliasLoader(self.alias_file)
        self.assertEqual(
            alias_interface.get_aliases(command=('ec2',)),
            {'list': 'describe-instances', 'regions': 'describe-regions'},
        )

    def test_get_aliases_multiple_nesting(self):
        with open(self.alias_file, 'a+') as f:
            f.write('[command ec2 wait]\n')
            f.write('instance-new-state = describe-instances\n')
        alias_interface = AliasLoader(self.alias_file)
        self.assertEqual(
            alias_interface.get_aliases(command=('ec2', 'wait')),
            {'instance-new-state': 'describe-instances'},
        )

    def test_command_aliases_cleanup_whitespace(self):
        with open(self.alias_file, 'a+') as f:
            f.write(
                '[command iam]\n'
                'my-alias = \n'
                '  my-alias-value \\ \n'
                '  --parameter foo\n'
            )
        alias_interface = AliasLoader(self.alias_file)
        self.assertEqual(
            alias_interface.get_aliases(command=('iam',)),
            # The backslash and newline should be preserved, but
            # still have the beginning and end newlines removed.
            {'my-alias': 'my-alias-value \\\n--parameter foo'},
        )


class TestAliasSubcommandInjector(unittest.TestCase):
    def setUp(self):
        self.files = FileCreator()
        self.alias_file = self.files.create_file('alias', '[command iam]\n')
        self.alias_loader = AliasLoader(self.alias_file)
        self.command_table = {}
        self.injector = AliasSubCommandInjector(self.alias_loader)
        self.command_object = mock.Mock(spec=CLICommand)
        self.session = mock.Mock(spec=Session)

    def tearDown(self):
        self.files.remove_all()

    def test_operation_alias_command(self):
        with open(self.alias_file, 'a+') as f:
            f.write('list = list-roles\n')

        self.injector.on_building_command_table(
            command_table=self.command_table,
            event_name='building-command-table.iam',
            session=self.session,
            command_object=self.command_object,
        )
        self.assertIn('list', self.command_table)
        self.assertIsInstance(
            self.command_table['list'], InternalAliasSubCommand
        )

    def test_global_parser_events_workflow(self):
        with open(self.alias_file, 'a+') as f:
            f.write('list = list-roles\n')

        clidriver = mock.Mock(spec=CLIDriver)
        fake_cmd_table = {'foo': mock.Mock(spec=CLICommand)}
        clidriver.subcommand_table = fake_cmd_table
        global_parser = mock.Mock()
        global_parser.parse_known_args.return_value = (FakeParsedArgs(), [])
        clidriver.create_parser.return_value = global_parser
        # First thing that fires building the main (i.e toplevel) command table.
        # This is the only inconsistent event args with respect to types where
        # the command object is a CLIDriver instead of a CLICommand subclass.
        self.injector.on_building_command_table(
            command_table=fake_cmd_table,
            event_name='building-command-table.main',
            session=self.session,
            command_object=clidriver,
        )
        # Then we fire the service specific command table event.
        self.injector.on_building_command_table(
            command_table=self.command_table,
            event_name='building-command-table.iam',
            session=self.session,
            command_object=self.command_object,
        )
        alias_cmd = self.command_table['list']
        alias_cmd([], FakeParsedArgs(command='iam'))
        # We should have plumbed through the global parser correctly:
        global_parser.parse_known_args.assert_called_with(
            ['iam', 'list-roles']
        )

    def test_noop_if_no_cmd_aliases_defined(self):
        self.injector.on_building_command_table(
            command_table=self.command_table,
            event_name='building-command-table.iam',
            session=self.session,
            command_object=self.command_object,
        )
        self.assertEqual(self.command_table, {})

    def test_can_inject_external_aliases_in_sub_cmd(self):
        with open(self.alias_file, 'a+') as f:
            f.write('list = !aws iam list-roles')

        self.injector.on_building_command_table(
            command_table=self.command_table,
            event_name='building-command-table.iam',
            session=self.session,
            command_object=self.command_object,
        )
        self.assertIn('list', self.command_table)
        self.assertIsInstance(self.command_table['list'], ExternalAliasCommand)


class TestAliasCommandInjector(unittest.TestCase):
    def setUp(self):
        self.files = FileCreator()
        self.alias_file = self.files.create_file('alias', '[toplevel]\n')
        self.alias_loader = AliasLoader(self.alias_file)
        self.session = mock.Mock(spec=Session)
        self.alias_cmd_injector = AliasCommandInjector(
            self.session, self.alias_loader
        )
        self.command_table = {}
        self.parser = MainArgParser(
            command_table=self.command_table,
            version_string='version',
            description='description',
            argument_table={},
        )

    def tearDown(self):
        self.files.remove_all()

    def test_service_alias_command(self):
        with open(self.alias_file, 'a+') as f:
            f.write('my-alias = my-alias-value\n')

        self.alias_cmd_injector.inject_aliases(self.command_table, self.parser)
        self.assertIn('my-alias', self.command_table)
        self.assertIsInstance(
            self.command_table['my-alias'], ServiceAliasCommand
        )

    def test_external_alias_command(self):
        with open(self.alias_file, 'a+') as f:
            f.write('my-alias = !my-alias-value\n')

        self.alias_cmd_injector.inject_aliases(self.command_table, self.parser)
        self.assertIn('my-alias', self.command_table)
        self.assertIsInstance(
            self.command_table['my-alias'], ExternalAliasCommand
        )

    def test_clobbers_builtins(self):
        builtin_cmd = mock.Mock(spec=CLICommand)
        self.command_table['builtin'] = builtin_cmd

        with open(self.alias_file, 'a+') as f:
            f.write('builtin = my-alias-value\n')

        self.alias_cmd_injector.inject_aliases(self.command_table, self.parser)
        self.assertIn('builtin', self.command_table)
        self.assertIsInstance(
            self.command_table['builtin'], ServiceAliasCommand
        )

    def test_shadow_proxy_command(self):
        builtin_cmd = mock.Mock(spec=CLICommand)
        builtin_cmd.name = 'builtin'
        self.command_table['builtin'] = builtin_cmd

        with open(self.alias_file, 'a+') as f:
            f.write('builtin = builtin\n')

        self.alias_cmd_injector.inject_aliases(self.command_table, self.parser)

        self.command_table['builtin']([], FakeParsedArgs(command='builtin'))
        # The builtin command should be passed to the alias
        # command when added to the table.
        builtin_cmd.assert_called_with([], FakeParsedArgs(command='builtin'))


class TestBaseAliasCommand(unittest.TestCase):
    def test_name(self):
        alias_cmd = BaseAliasCommand('alias-name', 'alias-value')
        self.assertEqual(alias_cmd.name, 'alias-name')
        alias_cmd.name = 'new-alias-name'
        self.assertEqual(alias_cmd.name, 'new-alias-name')


class TestInternalAliasSubCommand(unittest.TestCase):
    def setUp(self):
        self.alias_name = 'myalias'
        self.alias_value = 'list-stacks'
        self.session = mock.Mock(spec=Session)
        self.parent_command = 'cloudformation'
        self.command_object = mock.Mock(spec=CLICommand)
        self.command_object.name = self.parent_command
        self.shadow_proxy_command = None
        self.command_table = {
            self.parent_command: self.command_object,
        }
        self.global_args_parser = MainArgParser(
            command_table=self.command_table,
            version_string='version',
            description='description',
            argument_table={},
        )

    def create_alias_sub_command(self):
        return InternalAliasSubCommand(
            alias_name=self.alias_name,
            alias_value=self.alias_value,
            command_object=self.command_object,
            global_args_parser=self.global_args_parser,
            proxied_sub_command=self.shadow_proxy_command,
            session=self.session,
        )

    def test_alias_with_no_args(self):
        alias_cmd = self.create_alias_sub_command()
        self.assertEqual(alias_cmd.name, self.alias_name)
        parsed_globals = FakeParsedArgs(command=self.parent_command)
        alias_cmd(args=[], parsed_globals=parsed_globals)
        self.command_object.assert_called_with(
            [self.alias_value],
            parsed_globals,
        )

    def test_alias_with_alias_specific_args(self):
        # --stack-status-filter is specific to the list-stacks command
        # and is not a global param.
        self.alias_value = 'list-stacks --stack-status-filter CREATE_COMPLETE'
        alias_cmd = self.create_alias_sub_command()
        parsed_globals = FakeParsedArgs(command=self.parent_command)
        alias_cmd(args=[], parsed_globals=parsed_globals)
        self.command_object.assert_called_with(
            ['list-stacks', '--stack-status-filter', 'CREATE_COMPLETE'],
            parsed_globals,
        )

    def test_alias_with_global_args(self):
        self.alias_value = 'list-stacks --query StackSummaries[]'
        # The add_argument() calls configures the global_args_parser
        # to see the --query arg as a global arg and is properly moved
        # over to the set of parsed_globals.
        self.global_args_parser.add_argument('--query')
        alias_cmd = self.create_alias_sub_command()
        parsed_globals = FakeParsedArgs(command=self.parent_command)
        alias_cmd(args=[], parsed_globals=parsed_globals)
        self.command_object.assert_called_with(
            ['list-stacks'],
            parsed_globals,
        )
        # The --query arg should have moved to the parsed_globals
        self.assertEqual(parsed_globals.query, 'StackSummaries[]')

    def test_can_combine_global_args_alias_and_explicit(self):
        self.alias_value = 'list-stacks --query StackSummaries[]'
        # The add_argument() calls configures the global_args_parser
        # to see the --query arg as a global arg and is properly moved
        # over to the set of parsed_globals.
        self.global_args_parser.add_argument('--query')
        self.global_args_parser.add_argument('--endpoint-url')
        alias_cmd = self.create_alias_sub_command()
        parsed_globals = FakeParsedArgs(
            command=self.parent_command, endpoint_url='http://localhost:8000/'
        )
        alias_cmd(args=[], parsed_globals=parsed_globals)
        self.command_object.assert_called_with(
            ['list-stacks'],
            parsed_globals,
        )
        # The --query arg should have moved to the parsed_globals
        # and --endpoint-url should remain untouched.
        self.assertEqual(parsed_globals.query, 'StackSummaries[]')
        self.assertEqual(parsed_globals.endpoint_url, 'http://localhost:8000/')

    def test_explicit_value_overrides_alias_value_global_arg(self):
        self.alias_value = 'list-stacks --query StackSummaries[]'
        self.global_args_parser.add_argument('--query')
        alias_cmd = self.create_alias_sub_command()
        # A query value being in the FakeParsedArgs denotes
        # that a user explicitly provided this value on the command line,
        # e.g. 'aws cloudformation my-alias --query StackSummaries[].StackName'
        parsed_globals = FakeParsedArgs(
            command=self.parent_command,
            query='StackSummaries[].StackName',
        )
        alias_cmd(args=[], parsed_globals=parsed_globals)
        self.command_object.assert_called_with(
            ['list-stacks'],
            parsed_globals,
        )
        # The value from the alias definition overrides the explicitly provided
        # version.  This seems counterintuitive (I'd expect the more explicit
        # version from the command line args to take precedence) but this is the
        # existing behavior with top-level aliases so we want to ensure we preserve
        # this behavior with sub-command aliases.
        self.assertEqual(parsed_globals.query, 'StackSummaries[]')

    def test_alias_with_proxied_sub_command(self):
        self.shadow_proxy_command = mock.Mock(spec=CLICommand)
        self.alias_name = 'list-stacks'
        self.alias_value = 'list-stacks --stack-status-filter CREATE_COMPLETE'
        alias_cmd = self.create_alias_sub_command()
        parsed_globals = FakeParsedArgs(command=self.parent_command)
        alias_cmd(args=[], parsed_globals=parsed_globals)
        self.command_object.assert_not_called()
        self.shadow_proxy_command.assert_called_with(
            ['--stack-status-filter', 'CREATE_COMPLETE'],
            parsed_globals,
        )


class TestServiceAliasCommand(unittest.TestCase):
    def setUp(self):
        self.alias_name = 'myalias'
        self.session = mock.Mock(spec=Session)

    def create_command_table(self, services):
        command_table = {}
        for service in services:
            command_table[service] = mock.Mock(spec=CLICommand)
        return command_table

    def create_parser(self, command_table, extra_params=None):
        parser = MainArgParser(
            command_table=command_table,
            version_string='version',
            description='description',
            argument_table={},
        )
        if extra_params:
            for extra_param in extra_params:
                parser.add_argument('--' + extra_param)
        return parser

    def test_alias_with_only_service_command(self):
        alias_value = 'myservice'

        command_table = self.create_command_table([alias_value])
        parser = self.create_parser(command_table)
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        alias_cmd([], FakeParsedArgs(command=self.alias_name))
        command_table['myservice'].assert_called_with(
            [], FakeParsedArgs(command='myservice')
        )

    def tests_alias_with_shadow_proxy_command(self):
        alias_value = 'some-service'
        self.alias_name = alias_value

        shadow_proxy_command = mock.Mock(spec=CLICommand)
        shadow_proxy_command.name = alias_value

        command_table = {}
        parser = self.create_parser(command_table)

        alias_cmd = ServiceAliasCommand(
            self.alias_name,
            alias_value,
            self.session,
            command_table,
            parser,
            shadow_proxy_command,
        )
        command_table[self.alias_name] = alias_cmd

        alias_cmd([], FakeParsedArgs(command=self.alias_name))
        shadow_proxy_command.assert_called_with(
            [], FakeParsedArgs(command=alias_value)
        )

    def test_alias_with_shadow_proxy_command_no_match(self):
        alias_value = 'some-other-command'
        self.alias_name = 'some-service'

        shadow_proxy_command = mock.Mock(spec=CLICommand)
        shadow_proxy_command.name = 'some-service'

        command_table = self.create_command_table([alias_value])
        parser = self.create_parser(command_table)

        alias_cmd = ServiceAliasCommand(
            self.alias_name,
            alias_value,
            self.session,
            command_table,
            parser,
            shadow_proxy_command,
        )
        command_table[self.alias_name] = alias_cmd

        alias_cmd([], FakeParsedArgs(command=self.alias_name))
        command_table[alias_value].assert_called_with(
            [], FakeParsedArgs(command=alias_value)
        )
        # Even though it was provided, it should not be called because
        # the alias value did not match the shadow command name.
        self.assertFalse(shadow_proxy_command.called)

    def test_alias_with_operation_command(self):
        alias_value = 'myservice myoperation'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(command_table)
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        parsed_globals = FakeParsedArgs(command=self.alias_name)
        alias_cmd([], parsed_globals)
        command_table['myservice'].assert_called_with(
            ['myoperation'], FakeParsedArgs(command='myservice')
        )

    def test_alias_then_help_command(self):
        alias_value = 'myservice myoperation'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(command_table)
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        parsed_globals = FakeParsedArgs(command=self.alias_name)
        alias_cmd(['help'], parsed_globals)
        command_table['myservice'].assert_called_with(
            ['myoperation', 'help'], FakeParsedArgs(command='myservice')
        )

    def test_alias_then_additional_parameters(self):
        alias_value = 'myservice myoperation'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(command_table)
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        parsed_globals = FakeParsedArgs(command=self.alias_name)
        alias_cmd(['--parameter', 'val'], parsed_globals)
        command_table['myservice'].assert_called_with(
            ['myoperation', '--parameter', 'val'],
            FakeParsedArgs(command='myservice'),
        )

    def test_alias_with_operation_and_parameters(self):
        alias_value = 'myservice myoperation --my-parameter val'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(command_table)
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        alias_cmd([], FakeParsedArgs(command=self.alias_name))
        command_table['myservice'].assert_called_with(
            ['myoperation', '--my-parameter', 'val'],
            FakeParsedArgs(command='myservice'),
        )

    def test_alias_with_operation_and_global_parameters(self):
        alias_value = 'myservice myoperation --global-param val'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(
            command_table, extra_params=['global-param']
        )
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        alias_cmd([], FakeParsedArgs(command=self.alias_name))
        command_table['myservice'].assert_called_with(
            ['myoperation'],
            FakeParsedArgs(command='myservice', global_param='val'),
        )

    def test_maintains_global_defaults_when_missing_from_alias(self):
        alias_value = 'myservice myoperation'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(command_table)
        parser.add_argument('--global-with-default', default='default')
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )
        alias_cmd(
            [],
            FakeParsedArgs(
                command=self.alias_name, global_with_default='default'
            ),
        )
        command_table['myservice'].assert_called_with(
            ['myoperation'],
            FakeParsedArgs(command='myservice', global_with_default='default'),
        )

    def test_sets_global_parameters_when_differs_from_defaults(self):
        alias_value = 'myservice myoperation --global-with-default non-default'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(command_table)
        parser.add_argument('--global-with-default', default='default')
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        alias_cmd([], FakeParsedArgs(command=self.alias_name))
        command_table['myservice'].assert_called_with(
            ['myoperation'],
            FakeParsedArgs(
                command='myservice', global_with_default='non-default'
            ),
        )

    def test_global_parameters_can_be_emitted_and_modified(self):
        alias_value = 'myservice myoperation --global-param val'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(
            command_table, extra_params=['global-param']
        )
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        def replace_global_param_value_with_foo(event_name, **kwargs):
            parsed_args = kwargs['parsed_args']
            parsed_args.global_param = 'foo'

        self.session.emit.side_effect = replace_global_param_value_with_foo
        alias_cmd([], FakeParsedArgs(command=self.alias_name))
        self.session.emit.assert_called_with(
            'top-level-args-parsed', parsed_args=mock.ANY, session=self.session
        )

        command_table['myservice'].assert_called_with(
            ['myoperation'],
            FakeParsedArgs(command='myservice', global_param='foo'),
        )

    def test_properly_handles_multiple_spaces(self):
        alias_value = 'myservice myoperation    --my-parameter val'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(command_table)
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        alias_cmd([], FakeParsedArgs(command=self.alias_name))
        command_table['myservice'].assert_called_with(
            ['myoperation', '--my-parameter', 'val'],
            FakeParsedArgs(command='myservice'),
        )

    def test_properly_parses_aliases_broken_by_multiple_lines(self):
        alias_value = 'myservice myoperation \\' '\n--my-parameter val'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(command_table)
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        alias_cmd([], FakeParsedArgs(command=self.alias_name))
        command_table['myservice'].assert_called_with(
            ['myoperation', '--my-parameter', 'val'],
            FakeParsedArgs(command='myservice'),
        )

    def test_properly_preserves_quoted_values(self):
        alias_value = 'myservice myoperation --my-parameter \' \n$""\''

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(command_table)
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        alias_cmd([], FakeParsedArgs(command=self.alias_name))
        command_table['myservice'].assert_called_with(
            ['myoperation', '--my-parameter', ' \n$""'],
            FakeParsedArgs(command='myservice'),
        )

    def test_errors_when_service_command_is_invalid(self):
        alias_value = 'non-existent-service myoperation'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(command_table)
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        with self.assertRaises(ArgParseException):
            # Even though we catch the system exit, a message will always
            # be forced to screen because it happened at system exit.
            # The patch is to ensure it does not get displayed.
            with mock.patch('sys.stderr'):
                alias_cmd([], FakeParsedArgs(command=self.alias_name))

    def test_errors_when_no_service_command(self):
        alias_value = '--global-param=val'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(
            command_table, extra_params=['global-param']
        )
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        with self.assertRaises(ArgParseException):
            # Even though we catch the system exit, a message will always
            # be forced to screen because it happened at system exit.
            # The patch is to ensure it does not get displayed.
            with mock.patch('sys.stderr'):
                alias_cmd([], FakeParsedArgs(command=self.alias_name))

    def test_errors_when_shell_cannot_parse_alias(self):
        # Ending with a escape character that does not escape anything
        # is invalid and cannot be properly parsed.
        alias_value = 'myservice myoperation \\'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(command_table)
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        with self.assertRaises(InvalidAliasException):
            alias_cmd([], FakeParsedArgs(command=self.alias_name))

    def test_errors_when_unsupported_global_parameter_in_alias(self):
        # Unsupported global parameters are: --profile and --debug
        alias_value = 'myservice myoperation --profile value'

        command_table = self.create_command_table(['myservice'])
        parser = self.create_parser(command_table, extra_params=['profile'])
        alias_cmd = ServiceAliasCommand(
            self.alias_name, alias_value, self.session, command_table, parser
        )

        with self.assertRaises(InvalidAliasException):
            alias_cmd([], FakeParsedArgs(command=self.alias_name))


class TestExternalAliasCommand(unittest.TestCase):
    def setUp(self):
        self.subprocess_call = mock.Mock(spec=subprocess.call)

    def test_run_external_command(self):
        alias_value = '!ls'
        alias_cmd = ExternalAliasCommand(
            'alias-name', alias_value, invoker=self.subprocess_call
        )
        alias_cmd([], FakeParsedArgs(command='alias-name'))
        self.subprocess_call.assert_called_with('ls', shell=True)

    def test_external_command_returns_rc_of_subprocess_call(self):
        alias_value = '!ls'
        alias_cmd = ExternalAliasCommand(
            'alias-name', alias_value, invoker=self.subprocess_call
        )
        self.subprocess_call.return_value = 1
        self.assertEqual(
            alias_cmd([], FakeParsedArgs(command='alias-name')), 1
        )

    def test_external_command_uses_literal_alias_value(self):
        alias_value = '!f () {\n' '  ls .\n' '}; f'
        alias_cmd = ExternalAliasCommand(
            'alias-name', alias_value, invoker=self.subprocess_call
        )
        alias_cmd([], FakeParsedArgs(command='alias-name'))
        self.subprocess_call.assert_called_with(alias_value[1:], shell=True)

    def test_external_command_then_additional_args(self):
        alias_value = '!f () { ls "$1" }; f'
        alias_cmd = ExternalAliasCommand(
            'alias-name', alias_value, invoker=self.subprocess_call
        )
        alias_cmd(['extra'], FakeParsedArgs(command='alias-name'))
        self.subprocess_call.assert_called_with(
            'f () { ls "$1" }; f extra', shell=True
        )
